import logging

import numpy as np
import pytest

from egse.coordinates import ReferenceFrame
from egse.coordinates import dict_to_ref_model
from egse.coordinates import ref_model_to_dict
from egse.coordinates.refmodel import ReferenceFrameModel

LOGGER = logging.getLogger()


def test_construction():

    model = ReferenceFrameModel()

    assert len(model) == 0

    model = ReferenceFrameModel({"Master": ReferenceFrame.createMaster()})

    assert len(model) == 1

    # This should just be ignored and log a warning message.

    model.remove_frame("B")

    model_dict = model.serialize()

    assert "Master" in model_dict
    assert model_dict["Master"] == (
        "ReferenceFrame//([0.000000, 0.000000, 0.000000] | "
        "[0.000000, -0.000000, 0.000000] | Master | Master | [])"
    )

    model.add_frame("A", translation=[0.0, 0.0, 1.0], rotation=[0.0, 0.0, 0.0], ref="Master")

    with pytest.raises(KeyError):
        model.add_frame("A", translation=[0.0, 0.0, 1.0], rotation=[0.0, 0.0, 0.0], ref="Master")

    model.add_link("A", "Master")

    model_dict = model.serialize()

    print(model)

    assert len(model_dict) == len(model) == 2
    assert "A" in model_dict
    assert "A" in model


def test_construction_from_list():

    rot_config = "sxyz"
    master = ReferenceFrame.createMaster()

    a_ref = ReferenceFrame(
        transformation=np.identity(4), ref=master, name="a_ref", rot_config=rot_config
    )
    a_ref.addLink(master)

    b_ref = ReferenceFrame(
        transformation=np.identity(4), ref=a_ref, name="b_ref", rot_config=rot_config
    )

    c_ref = ReferenceFrame.fromTranslationRotation(
        [-2, -2, -2], [-3, -4, -5], rot_config=rot_config, ref=b_ref, name="c_ref"
    )
    c_ref.addLink(b_ref)
    c_ref.addLink(a_ref)

    d_ref = ReferenceFrame(
        transformation=np.identity(4), ref=c_ref, name="d_ref", rot_config=rot_config
    )

    model_list = [a_ref, b_ref, c_ref, d_ref, master]

    model = ReferenceFrameModel(model_list)

    assert "a_ref" in model
    assert "b_ref" in model
    assert "c_ref" in model
    assert "Master" in model

    print(model)


def test_add_frame_from_transformation():

    model = ReferenceFrameModel()

    model.add_master_frame()
    model.add_frame("a_ref", transformation=np.identity(4), ref="Master")
    model.add_link("a_ref", "Master")

    print(model)


def test_move_absolute():

    model = ReferenceFrameModel()

    model.add_master_frame()
    model.add_frame("A", translation=[4, 0, 2], rotation=[0, 0, 0], ref="Master")
    model.add_frame("B", translation=[-2, 0, 7], rotation=[0, 0, 0], ref="Master")

    print(model)

    model.move_absolute_self("A", [1, 0, 1], [0, 0, 0])

    print(model)

    model.add_link("A", "B")

    model.move_absolute_self("A", [-1, 0, 1], [0, 0, 0])

    print(model)

    # FIXME:
    #   This should not be possible, because 'A' and 'B' are linked.

    model.move_absolute_in_other("B", "A", [1, 0, 1], [0, 0, 0])

    print(model)


def test_proper_ref_model():
    rot_config = "sxyz"
    master = ReferenceFrame.createMaster()

    a_ref = ReferenceFrame(transformation=np.identity(4), ref=master, name="a_ref",
                           rot_config=rot_config)
    a_ref.addLink(master)

    b_ref = ReferenceFrame(transformation=np.identity(4), ref=a_ref, name="b_ref",
                           rot_config=rot_config)

    c_ref = ReferenceFrame.fromTranslationRotation(
        [-2, -2, -2], [-3, -4, -5], rot_config=rot_config, ref=b_ref, name="c_ref"
    )
    c_ref.addLink(b_ref)
    c_ref.addLink(a_ref)

    d_ref = ReferenceFrame(transformation=np.identity(4), ref=c_ref, name="d_ref",
                           rot_config=rot_config)

    model_list = [a_ref, b_ref, c_ref, d_ref, master]

    csl_model = ref_model_to_dict(model_list)
    LOGGER.info(csl_model.pretty_str())

    LOGGER.info(csl_model["a_ref"])
    LOGGER.info(csl_model.a_ref)

    ref_model = dict_to_ref_model(csl_model)

    LOGGER.info(ref_model.pretty_str())

    LOGGER.info(ref_model["a_ref"])
    LOGGER.info(ref_model.a_ref)

    LOGGER.info(ref_model.a_ref.find_master())

    # the model has no master

    no_master_list = [a_ref, b_ref, c_ref, d_ref]

    # check if there is any reference frame for which the master is not in the model

    x = [ref.find_master() in no_master_list for ref in no_master_list]
    LOGGER.warning(x)

    # find any reference frame that refers to another reference frame not inside the model

    x = [ref for ref in no_master_list if ref.ref not in no_master_list]
    LOGGER.warning(x)

    # the model is missing a reference frame

    missing_ref_model = [master, b_ref, d_ref]

    x = [ref for ref in missing_ref_model if ref.ref not in missing_ref_model]
    LOGGER.warning(x)



import logging

from egse.control import ControlServer
from egse.device import DeviceTimeoutError
from egse.hk import read_conversion_dict
from egse.metrics import define_metrics
from egse.protocol import CommandProtocol
from egse.settings import Settings
from egse.setup import load_setup
from egse.synoptics import SynopticsManagerProxy
from egse.system import format_datetime
from egse.tempcontrol.keithley.daq6510 import DAQ6510Controller
from egse.tempcontrol.keithley.daq6510 import DAQ6510Interface
from egse.tempcontrol.keithley.daq6510 import DAQ6510Simulator
from egse.tempcontrol.keithley.daq6510_dev import DAQ6510Command
from egse.zmq_ser import bind_address

COMMAND_SETTINGS = Settings.load(filename="daq6510.yaml")

MODULE_LOGGER = logging.getLogger(__name__)


class DAQ6510Protocol(CommandProtocol):
    def __init__(self, control_server: ControlServer):
        """Initialisation of a new Protocol for DAQ6510 Management.

        Args:
            control_server: Control Server for which to send out status and monitoring information
        """

        super().__init__()
        self.control_server = control_server

        if Settings.simulation_mode():
            self.daq = DAQ6510Simulator()
        else:
            self.daq = DAQ6510Controller()

        try:
            self.daq.connect()
        except (ConnectionError, DeviceTimeoutError):
            MODULE_LOGGER.warning("Couldn't establish a connection to the DAQ6510, check the log messages.")

        self.load_commands(COMMAND_SETTINGS.Commands, DAQ6510Command, DAQ6510Interface)
        self.build_device_method_lookup_table(self.daq)

        setup = load_setup()
        self.channels = setup.gse.DAQ6510.channels

        self.hk_conversion_table = read_conversion_dict(self.control_server.get_storage_mnemonic(), setup=setup)

        self.synoptics = SynopticsManagerProxy()
        self.metrics = define_metrics(origin="DAS-DAQ6510", use_site=True, setup=setup)

    def get_bind_address(self) -> str:
        """Returns a string with the bind address, the endpoint, for accepting connections and bind a socket to.


        Returns: String with the protocol and port to bind a socket to.
        """

        return bind_address(
            self.control_server.get_communication_protocol(),
            self.control_server.get_commanding_port(),
        )

    def get_status(self) -> dict:
        """Returns a dictionary with status information for the Control Server and the DAQ6510.

        Returns: Dictionary with status information for the Control Server and the DAQ6510.
        """

        return super().get_status()

    def get_housekeeping(self) -> dict:
        """Returns a dictionary with housekeeping information about the DAQ6510.

        Returns: Dictionary with housekeeping information about the DAQ6510.
        """

        hk_dict = dict()
        hk_dict["timestamp"] = format_datetime()

        # # TODO I guess we have to do something along those lines
        # # (We'll have to increase the HK delay, cfr. Agilents)
        # measurement = self.daq.perform_measurement(channel_list=self.channels)
        # temperatures = convert_hk_names(measurement, self.hk_conversion_table)
        # hk_dict.update(temperatures)
        #
        # self.synoptics.store_th_synoptics(hk_dict)
        #
        # for key, value in hk_dict.items():
        #     if key != "timestamp":
        #         self.metrics[key].set(value)

        return hk_dict

    def quit(self) -> None:
        """Clean up and stop threads that were started by the process."""

        # TODO
        # self.synoptics.disconnect_cs()
        pass



from pathlib import Path

from PyQt5.QtWidgets import QFrame
from PyQt5.QtWidgets import QMessageBox

from egse.resource import initialise_resources

# Load the resources that are defined by this package, we do not have resource ids other than those
# defined by default, therefore no need to call add_resource_id().

initialise_resources(Path(__file__).parent)


def show_warning_message(description: str, info_text: str = None, detailed_text: str = None):
    msg_box = QMessageBox()

    msg_box.setWindowTitle("Warning")
    msg_box.setIcon(QMessageBox.Warning)
    msg_box.setText(description)
    if info_text is not None:
        msg_box.setInformativeText(info_text)
    if detailed_text is not None:
        msg_box.setDetailedText(detailed_text)
    msg_box.setStandardButtons(QMessageBox.Ok)
    msg_box.exec_()


def show_info_message(description: str, info_text: str = None, detailed_text: str = None):
    msg_box = QMessageBox()

    msg_box.setWindowTitle("Warning")
    msg_box.setIcon(QMessageBox.Information)
    msg_box.setText(description)
    if info_text is not None:
        msg_box.setInformativeText(info_text)
    if detailed_text is not None:
        msg_box.setDetailedText(detailed_text)
    msg_box.setStandardButtons(QMessageBox.Ok)
    msg_box.exec_()


class VLine(QFrame):
    """Presents a simple Vertical Bar that can be used in e.g. the status bar."""

    def __init__(self):
        super().__init__()
        self.setFrameShape(self.VLine | self.Sunken)


class QHLine(QFrame):
    """Presents a simple Horizontal Bar that can be used in e.g. the status bar."""

    def __init__(self):
        super(QHLine, self). __init__()
        self.setFrameShape(QFrame.HLine)
        self.setFrameShadow(QFrame.Sunken)



"""
The Reference Frame Model is a self-consistent graph of reference frames that are connected by
reference and/or by explicit links.

The ReferenceFrameModel class will keep the model consistent and self-contained.

Functionality:

* Initialization from a list of ReferenceFrames.
* Serialization into a dictionary that can be saved into a YAML file.
* Manipulation of the model
    * Move a reference frame (translation, rotation) with respect to another reference frame
        * Absolute movement, center of rotation either local or other
        * Relative Movement, center of rotation either local or other
    * Move a reference frame (translation, rotation) with respect to itself
        * Absolute movement
        * Relative Movement
    * Change the definition of a reference frame in the model
* Inspection of the model
    * Get the definition of a reference frame (what should this be? only translation & rotation?)
    * Get position of a reference frame
    * Get the position of a point in a 'target' reference frame, but defined in a 'source'
      reference frame
    * Get a string representation of the model.
    * Find inconsistencies in the model
    * What other information do we need?
        * find the path from one reference frame to another reference frame?
        * find all reference frames that are affected by a movement or redefinition of a
          reference frame?
        * ...

"""
import logging
from typing import Dict
from typing import List
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.transforms import ScaledTranslation
from mpl_toolkits.mplot3d import Axes3D

import egse.coordinates.transform3d_addon as t3add
from egse.coordinates import dict_to_ref_model
from egse.coordinates import ref_model_to_dict
from egse.coordinates.referenceFrame import ReferenceFrame
from egse.setup import NavigableDict

LOGGER = logging.getLogger(__name__)

# TODO : HANDLING "moving_in_ref" (obusr)  in move_absolute_ext and move_relative_ext
#        should it be added to the model temporarily ??
#        after the move : remove the link and delete that frame
#
# Priority 1
#   * access methods to allow for things like is_avoidance_ok(hexsim.cs_user, hexsim.cs_object,
#     setup=setup, verbose=True)
#
#  Priority 2
#   * Move name handling from ReferenceFrame to here (if necessary)


class ReferenceFrameModel:
    """
    A ReferenceFrameModel is a collection of reference frames that are linked to each other to
    form a Graph.
    """

    _ROT_CONFIG_DEFAULT = "sxyz"
    _ACTIVE_DEFAULT = True
    _DEGREES_DEFAULT = True

    def __init__(
        self,
        model: Union[Dict, List[ReferenceFrame]] = None,
        rot_config: str = _ROT_CONFIG_DEFAULT,
        use_degrees: bool = _DEGREES_DEFAULT,
        use_active_movements: bool = _ACTIVE_DEFAULT,
    ):
        """
        When the model_dict is empty or None, a new model is created with a master reference frame.

        Args:
            model: a list or a dictionary of reference frames that make up the model
            use_degrees: use degrees throughout this model unless explicitly specified in the
                function call.
        """
        self._use_degrees = use_degrees
        self._use_active_movements = use_active_movements
        self._rot_config = rot_config

        # Keep a dictionary with all reference frames that are part of the model. The keys shall
        # be the name of the reference frame. When the model passed is empty, create only a
        # master reference frame.

        if isinstance(model, (dict, list)):
            self._model = self.deserialize(model)
        else:
            self._model = NavigableDict({})

    def __str__(self):
        return self._model.pretty_str()

    def __len__(self):
        return len(self._model)

    def __contains__(self, item):
        return item in self._model

    def __iter__(self):
        return iter(self._model.values())

    def summary(self):
        result = f"Nb of frames: {len(self)}\n"

        for ref in self:
            result += (
                f"{ref.name:>10}[{ref.ref.name}]" f"  ---  {[link.name for link in ref.linkedTo]}\n"
            )
        return result

    @staticmethod
    def deserialize(model_dict: dict) -> NavigableDict:
        """
        Deserialize means you take a serialized representation of a reference frames model and
        turn it into a dictionary containing all the reference frames with their links and
        references.

        Args:
            model_dict: a dictionary of serialized reference frames

        Returns:
            A dictionary of ReferenceFrames that make up a model.

        """
        return dict_to_ref_model(model_dict)

    def serialize(self) -> NavigableDict:
        """
        Serialize the model by serializing each of the reference frames into an object that can
        easily be saved to a YAML or a JSON file. Return a dictionary with the serialized frames.

        Returns:
            A dictionary with all the reference framed serialized.
        """

        return ref_model_to_dict(self._model)

    def add_master_frame(self):

        # TODO: First check if there is not already a Master frame in the model

        self._model["Master"] = ReferenceFrame.createMaster()

    def add_frame(
        self,
        name: str,
        *,
        translation: List[float] = None,
        rotation: List[float] = None,
        transformation=None,
        ref: str,
    ):
        """
        Add a reference frame to the model.

        .. note::
            Only the `name` parameter can be positional, all the other arguments (translation,
            rotation, transformation, and ref) must be given as keyword arguments.

        Args:
            name: the name for the reference frame. This name is it's identifier within the model.
            translation: the translation vector
            rotation: the rotation vector
            transformation: the transformation vector, if `transformation` is given,
                both `translation` and `rotation` are ignored.
            ref: the reference frame that is a reference for 'name', i.e. 'name' is defined with
                respect to 'ref'.
        """

        if name in self._model:
            raise KeyError("A reference frame with the name '{name} already exists in the model.")

        ref = self._model[ref]

        if transformation is not None:
            self._model[name] = ReferenceFrame(
                transformation, ref=ref, name=name, rot_config=self._rot_config,
            )
        else:
            self._model[name] = ReferenceFrame.fromTranslationRotation(
                translation,
                rotation,
                name=name,
                ref=ref,
                rot_config=self._rot_config,
                degrees=self._use_degrees,
                active=self._use_active_movements,
            )

    def remove_frame(self, name: str):
        """
        Deletes the reference frame from the model. If the reference frame doesn't exist in the
        model, a warning message is logged.

        Args:
            name: the name of the reference frame to remove
        """

        if name in self._model:

            frame: ReferenceFrame = self._model[name]

            # We need to get the links out in a list because the frame.removeLink() method deletes
            # frames from the linkedTo dictionary and that is not allowed in a for loop.

            links = [linked_frame for linked_frame in frame.linkedTo]
            for link in links:
                frame.removeLink(link)

            del self._model[name]
        else:
            LOGGER.warning(
                f"You tried to remove a non-existing reference frame '{name}' from the model."
            )

    def get_frame(self, name: str) -> ReferenceFrame:
        """
        Returns a frame with the given name.

        .. note::
            Use this function with care since this breaks encapsulation and may lead to an
            inconsistent model when the frame is changed outside of the scope of the reference
            model.

        Args:
            name: the name of the requested reference frame

        Returns:
            The reference frame with the given name.
        """
        return self._model[name]

    def add_link(self, source: str, target: str):
        """
        Add a link between two reference frames. All links are bi-directional.

        Args:
            source: the source reference frame
            target: the target reference frame

        """
        if source not in self._model:
            raise KeyError(f"There is no reference frame with the name '{source} in the model.")
        if target not in self._model:
            raise KeyError(f"There is no reference frame with the name '{target} in the model.")

        source = self._model[source]
        target = self._model[target]

        source.addLink(target)

    def remove_link(self, source: str, target: str):
        """
        Remove a link between two reference frames. All links are bi-directional and this method
        removes both links.

        Args:
            source: the source reference frame
            target: the target reference frame

        """
        if source not in self._model:
            raise KeyError(f"There is no reference frame with the name '{source} in the model.")
        if target not in self._model:
            raise KeyError(f"There is no reference frame with the name '{target} in the model.")

        source = self._model[source]
        target = self._model[target]

        source.removeLink(target)

    def move_absolute_self(self, frame: str, translation, rotation, degrees=_DEGREES_DEFAULT):
        """
        Apply an absolute movement to the given ReferenceFrame such that it occupies a given
        absolute position wrt "frame_ref" after the movement.

        NO Hexapod equivalent.

        Args:
            frame (str): the name of the reference frame to move
        """

        frame = self._model[frame]
        frame.setTranslationRotation(
            translation,
            rotation,
            rot_config=self._rot_config,
            active=self._use_active_movements,
            degrees=degrees,
            preserveLinks=True,
        )

    def move_absolute_in_other(
        self, frame: str, other: str, translation, rotation, degrees=_DEGREES_DEFAULT
    ):
        """
        Apply an absolute movement to the ReferenceFrame "frame", such that it occupies
        a given absolute position with respect to "other" after the movement.

        EQUIVALENT PunaSimulator.move_absolute, setting hexobj wrt hexusr.

        Args:
            frame (str): the name (id) of the reference frame to move
            other (str): the name (id) of the reference frame
        """

        # TODO:
        #   There can not be a link between frame and other, not direct and not indirect.
        #   So, with A-link-B-link-C-link-D, we can not do move_absolute_in_other('A', 'D', ...)

        frame = self._model[frame]
        other = self._model[other]

        transformation = other.getActiveTransformationTo(frame)

        moving_in_other = ReferenceFrame(
            transformation, rot_config=self._rot_config, ref=other, name="moving_in_other"
        )

        moving_in_other.addLink(frame)

        moving_in_other.setTranslationRotation(
            translation,
            rotation,
            rot_config=self._rot_config,
            active=self._use_active_movements,
            degrees=degrees,
            preserveLinks=True,
        )

        moving_in_other.removeLink(frame)

        del moving_in_other

    def move_relative_self(self, frame: str, translation, rotation, degrees=_DEGREES_DEFAULT):
        """
        Apply a relative movement to the given ReferenceFrame assuming the movement is expressed
        in that same frame.

        EQUIVALENT PunaSimulator.move_relative_object

        Args:
            frame (str): the name of the reference frame to move
        """

        frame = self._model[frame]
        frame.applyTranslationRotation(
            translation,
            rotation,
            rot_config=self._rot_config,
            active=self._use_active_movements,
            degrees=degrees,
            preserveLinks=True,
        )

    def move_relative_other(
        self, frame: str, other: str, translation, rotation, degrees=_DEGREES_DEFAULT
    ):
        """
        Apply a relative movement to the ReferenceFrame "frame". The movement is expressed wrt
        the axes of another frame, "other".

        The center of rotation is the origin of the reference frame 'other'.

        NO Hexapod equivalent.

        Args:
            frame (str): the name (id) of the reference frame to move
            other (str): the name (id) of the reference frame
        """

        # TODO:
        #   There can not be a link between frame and other, not direct and not indirect.
        #   So, with A-link-B-link-C-link-D, we can not do move_absolute_in_other('A', 'D', ...)

        frame = self._model[frame]
        other = self._model[other]

        transformation = frame.getActiveTransformationTo(other)

        moving_in_other = ReferenceFrame(
            transformation, rot_config=self._rot_config, ref=other, name="moving_in_other"
        )

        moving_in_other.addLink(frame)

        moving_in_other.applyTranslationRotation(
            translation,
            rotation,
            rot_config=self._rot_config,
            active=self._use_active_movements,
            degrees=degrees,
            preserveLinks=True,
        )

        moving_in_other.removeLink(frame)

        del moving_in_other  # not need as local scope

    def move_relative_other_local(
        self, frame: str, other: str, translation, rotation, degrees=_DEGREES_DEFAULT
    ):
        """
        Apply a relative movement to the ReferenceFrame "frame".

        The movement is expressed wrt the axes of an external frame "other"

        The center of rotation is the origin of the reference frame 'frame'.

        EQUIVALENT PunaSimulator.move_relative_user

        """

        # TODO:
        #   There can not be a link between frame and other, not direct and not indirect.
        #   So, with A-link-B-link-C-link-D, we can not do move_absolute_in_other('A', 'D', ...)

        frame = self._model[frame]
        other = self._model[other]

        # Represent the requested movement

        # Derotation of MOVING --> REF  (align frame_moving axes on those of frame_ref)

        derotation = frame.getActiveTransformationTo(other)
        derotation[:3, 3] = [0, 0, 0]

        # Reverse rotation (record inverse rotation, to restore the frame in the end)

        rerotation = derotation.T

        # Requested translation matrix  (already expressed wrt frame_ref)

        translation_ = np.identity(4)
        translation_[:3, 3] = translation

        # Requested rotation matrix (already expressed wrt frame_ref)

        zeros = [0, 0, 0]
        rotation_ = t3add.translationRotationToTransformation(
            zeros, rotation, rot_config=self._rot_config
        )

        # All translations and rotations are applied to frame_moving
        # ==> a. need for "derotation" before applying the translation
        #     b. the center or rotation is always the origin of frame_moving
        # 1. rotate frame_moving to align it with frame_ref (i.e. render their axes parallel)
        # 2. apply the translation in this frame
        # 3. restore the original orientation of the moving frame
        # 4. apply the requested rotation

        transformation = derotation @ translation_ @ rerotation @ rotation_

        # Apply the requested movement

        frame.applyTransformation(transformation, preserveLinks=True)


def plot_ref_model(model: ReferenceFrameModel):

    # figsize is in inch, 6 inch = 15.24 cm, 5 inch = 12.7 cm

    fig = plt.figure(figsize=(6, 5), dpi=100)

    ax = fig.add_subplot(1, 1, 1)

    # Set axes limits in data coordinates

    ax.set_xlim(-10, 10)
    ax.set_ylim(-10, 10)
    ax.set_xticks(range(-10, 11, 2))
    ax.set_yticks(range(-10, 11, 2))
    ax.grid(True)

    for frame in model:
        draw_frame(ax, frame, plane="xz")

    plt.show()


def plot_ref_model_3d(model: ReferenceFrameModel):

    fig = plt.figure(figsize=(8, 8), dpi=100)
    ax = Axes3D(fig)
    # ax.set_box_aspect([1, 1, 1])

    ax.set_xlabel("x")
    ax.set_ylabel("y")
    ax.set_zlabel("z")

    def get_fix_mins_maxs(mins, maxs):
        deltas = (maxs - mins) / 12.0
        mins = mins + deltas / 4.0
        maxs = maxs - deltas / 4.0

        return [mins, maxs]

    min_ = -8
    max_ = 8
    minmax = get_fix_mins_maxs(min_, max_)

    # ax.set_xticks(range(min_, max_, 2))
    # ax.set_yticks(range(min_, max_, 2))
    # ax.set_zticks(range(min_, max_, 2))

    ax.set_xlim(minmax)
    ax.set_ylim(minmax)
    ax.set_zlim(minmax)

    delta = 0.1
    ax.scatter(
        [min_ + delta, max_ - delta],
        [min_ + delta, max_ - delta],
        [min_ + delta, max_ - delta],
        color="k",
        marker=".",
    )

    for frame in model:
        draw_frame_3d(ax, frame)

    # ax.set_proj_type('ortho')
    ax.set_proj_type("persp")

    set_axes_equal(ax)
    plt.show()


# The aspect ration of the plots is not equal by default.
# This solution was given in SO: https://stackoverflow.com/a/63625222/4609203


def set_axes_equal(ax: plt.Axes):
    """Set 3D plot axes to equal scale.

    Make axes of 3D plot have equal scale so that spheres appear as
    spheres and cubes as cubes.  Required since `ax.axis('equal')`
    and `ax.set_aspect('equal')` don't work on 3D.
    """
    limits = np.array([ax.get_xlim3d(), ax.get_ylim3d(), ax.get_zlim3d()])
    origin = np.mean(limits, axis=1)
    radius = 0.5 * np.max(np.abs(limits[:, 1] - limits[:, 0]))
    x, y, z = origin
    ax.set_xlim3d([x - radius, x + radius])
    ax.set_ylim3d([y - radius, y + radius])
    ax.set_zlim3d([z - radius, z + radius])


def draw_frame_3d(ax: Axes3D, frame: ReferenceFrame, DEFAULT_AXIS_LENGTH=100, **kwargs):

    master = frame.find_master()

    f0 = frame.getOrigin()
    fx = frame.getAxis("x", name="fx")
    fy = frame.getAxis("y", name="fy")
    fz = frame.getAxis("z", name="fz")
    f0m = f0.expressIn(master)[:3]
    fxm = fx.expressIn(master)[:3]
    fym = fy.expressIn(master)[:3]
    fzm = fz.expressIn(master)[:3]

    # Origin of the X,Y and Z vectors (x = the 'x' coordinates of the origin of all 3 vectors)
    # Every vector independently (--> plot in diff. colors)
    x, y, z = np.array([f0m[0]]), np.array([f0m[1]]), np.array([f0m[2]])

    # Orientation of the X,Y and Z vectors
    vecxx, vecyx, veczx = (
        np.array([fxm[0] - f0m[0]]),
        np.array([fym[0] - f0m[0]]),
        np.array([fzm[0] - f0m[0]]),
    )
    vecxy, vecyy, veczy = (
        np.array([fxm[1] - f0m[1]]),
        np.array([fym[1] - f0m[1]]),
        np.array([fzm[1] - f0m[1]]),
    )
    vecxz, vecyz, veczz = (
        np.array([fxm[2] - f0m[2]]),
        np.array([fym[2] - f0m[2]]),
        np.array([fzm[2] - f0m[2]]),
    )

    kwargs.setdefault("length", 2)
    kwargs.setdefault("normalize", True)

    ax.quiver(x, y, z, vecxx, vecxy, vecxz, color="r", **kwargs)
    ax.quiver(x, y, z, vecyx, vecyy, vecyz, color="g", **kwargs)
    ax.quiver(x, y, z, veczx, veczy, veczz, color="b", **kwargs)

    offset = 0.1
    ax.text(f0m[0] + offset, f0m[1] + offset, f0m[2] + offset, frame.name)


def draw_frame(ax, frame: ReferenceFrame, plane="xz", DEFAULT_AXIS_LENGTH=100):

    fig = ax.get_figure()

    # FC : Figure coordinates (pixels)
    # NFC : Normalized figure coordinates (0 → 1)
    # DC : Data coordinates (data units)
    # NDC : Normalized data coordinates (0 → 1)

    dc2fc = ax.transData.transform
    fc2dc = ax.transData.inverted().transform
    fc2ndc = ax.transAxes.inverted().transform

    def dc2ndc(x):  # better than defining and assigning a lambda function
        return fc2ndc(dc2fc(x))

    x_idx, y_idx = {"xz": (0, 2), "xy": (0, 1), "yz": (1, 2)}[plane]

    # Draw the origin

    origin = frame.getOrigin()
    origin_in_master = origin.expressIn(frame.find_master())

    ax.scatter([origin_in_master[x_idx]], [origin_in_master[y_idx]], color="k")

    # Draw the axis

    origin_dc = np.array([[origin_in_master[x_idx], origin_in_master[y_idx]]])

    point = dc2fc(origin_dc[0])
    point[0] += DEFAULT_AXIS_LENGTH
    target_dc = np.append(origin_dc, [fc2dc(point)], axis=0)

    ax.plot(target_dc[:, 0], target_dc[:, 1], color="k")

    point = dc2fc(origin_dc[0])
    point[1] += DEFAULT_AXIS_LENGTH
    target_dc = np.append(origin_dc, [fc2dc(point)], axis=0)

    ax.plot(target_dc[:, 0], target_dc[:, 1], color="k")

    # Draw the axes label

    dx, dy = 10 / fig.dpi, 10 / fig.dpi
    offset = ScaledTranslation(dx, dy, fig.dpi_scale_trans)
    point = dc2ndc(origin_dc[0])
    plt.text(point[0], point[1], frame.name, transform=ax.transAxes + offset)


def define_the_initial_setup():
    model = ReferenceFrameModel()

    model.add_master_frame()
    model.add_frame("A", translation=[2, 2, 2], rotation=[0, 0, 0], ref="Master")
    model.add_frame("B", translation=[-2, 2, 2], rotation=[0, 0, 0], ref="Master")
    model.add_frame("C", translation=[2, 2, 5], rotation=[0, 0, 0], ref="A")
    model.add_frame("D", translation=[2, 2, 2], rotation=[0, 0, 0], ref="B")

    model.add_link("A", "B")
    model.add_link("B", "C")

    print(model.serialize())
    plot_ref_model_3d(model)

    return model


def get_vectors(rf1, rf2, model):
    """
    get_vectors(rf1,rf2, model)
    :param rf1: string : name of ref. frame "from"
    :param rf2: string : name of ref. frame "to"
    :param model: CSLReferenceFrameModel containing rf1 and rf2
    :return: translation and rotation vectors from rf1 to rf2
    """
    return model.get_frame(rf1).getActiveTranslationRotationVectorsTo(model.get_frame(rf2))


def print_vectors(rf1, rf2, model):
    """
    :param rf1: string : name of ref. frame "from"
    :param rf2: string : name of ref. frame "to"
    :param model: CSLReferenceFrameModel containing rf1 and rf2
    :return: N.A.
    Prints the translation and rotation vectors from rf1 to rf2
    """
    trans, rot = model.get_frame(rf1).getActiveTranslationRotationVectorsTo(model.get_frame(rf2))
    print(f"{rf1:8s} -> {rf2:8s} : Trans [{trans[0]:11.4e}, {trans[1]:11.4e}, {trans[2]:11.4e}]    Rot [{rot[0]:11.4e}, {rot[1]:11.4e}, {rot[2]:11.4e}]")
    return


if __name__ == "__main__":

    logging.basicConfig(level=20)

    model = define_the_initial_setup()

    print(model.summary())

    print("\nMove frame 'A', frames 'B' and 'C' move with it.\n")
    model.move_absolute_self("A", [1, 1, 3], [0, 0, 45])
    print(model.serialize())
    plot_ref_model_3d(model)

    model = define_the_initial_setup()

    print("\nMove frame 'B' with respect to 'Master, frames 'A' and 'C' move with it.\n")
    model.move_absolute_in_other("B", "Master", [1, 1, -1], [0, 0, 0])
    print(model.serialize())
    plot_ref_model_3d(model)

    model = define_the_initial_setup()

    print("\nMove frame 'D' relative to itself, turn 45º\n")
    model.move_relative_self("D", [0, 0, 0], [45, 0, 0])
    print(model.serialize())
    plot_ref_model_3d(model)

    model = define_the_initial_setup()

    print("\nMove frame 'D' relative to 'A', turn 45º around origin of 'A'\n")
    model.move_relative_other("D", "A", [0, 0, 0], [0, 45, 0])
    print(model.serialize())
    plot_ref_model_3d(model)

    model = define_the_initial_setup()

    print("\nMove frame 'D' relative to 'A', turn 45º around origin of 'D'\n")
    model.move_relative_other_local("D", "A", [0, 0, 0], [0, 45, 0])
    print(model.serialize())
    plot_ref_model_3d(model)

    model = define_the_initial_setup()



"""
This module provides Process Management functionality for the Common-EGSE.
"""
import importlib
import logging
import pickle
import time

import sys
from pathlib import Path

import zmq

from egse.command import ClientServerCommand
from egse.confman import ConfigurationManagerProxy
from egse.control import ControlServer
from egse.response import Success, Failure
from egse.decorators import dynamic_interface
from egse.dpu import fitsgen
from egse.dpu.dpu_cs import is_dpu_cs_active
from egse.fee import n_fee_hk
from egse.fov import fov_hk
from egse.process import SubProcess
from egse.protocol import CommandProtocol
from egse.proxy import Proxy
from egse.settings import Settings
from egse.setup import Setup
from egse.state import GlobalState
from egse.storage import is_storage_manager_active
from egse.system import find_class
from egse.system import format_datetime
from egse.zmq_ser import bind_address
from egse.zmq_ser import connect_address

HERE = Path(__file__).parent

CTRL_SETTINGS = Settings.load("Process Manager Control Server")
COMMAND_SETTINGS = Settings.load(location=HERE, filename="procman.yaml")

LOGGER = logging.getLogger(__name__)


def is_process_manager_cs_active(timeout: float = 0.5):
    """Checks if the Process Manager Control Server is active.

    To check whether the Control Server is active, a "Ping" command is sent.
    If a "Pong" reply is received before timeout, the Control Server is said
    to be active (and True will be returned).  If no reply is received before
    timeout or if the reply is not "Pong", the Control Server is said to be
    inactive (and False will be returned).

    Args:
        - timeout (float): Timeout when waiting for a reply [s].

    Returns:
        True if the Process Manager Control Server is active; False otherwise.
    """

    # Create a socket and connect it to the commanding port of the CS

    socket = zmq.Context.instance().socket(zmq.REQ)
    endpoint = connect_address(
        CTRL_SETTINGS.PROTOCOL, CTRL_SETTINGS.HOSTNAME, CTRL_SETTINGS.COMMANDING_PORT
    )
    socket.connect(endpoint)

    # Send a "Ping" command and wait for a reply
    # (but not beyond timeout)

    data = pickle.dumps("Ping")
    socket.send(data)
    rlist, _, _ = zmq.select([socket], [], [], timeout=timeout)

    # Reply received before timeout
    # (should be "Pong")

    status = False

    if socket in rlist:

        # Only if the reply is "Pong", the CS is active

        data = socket.recv()
        response = pickle.loads(data)

        status = response == "Pong"

    # No reply received -> inactive

    socket.close(linger=0)

    return status


class ProcessManagerProtocol(CommandProtocol):

    """
    Command Protocol for Process Management.
    """

    def __init__(self, control_server: ControlServer):
        """Initialisation of a new Protocol for Process Management.

        The initialisation of this Protocol consists of the following steps:

            - create a Controller to which the given Control Server should send commands;
            - load the commands;
            - build a look-up table for the commands.

        Args:
            - control_server: Control Server via which commands should be sent
                              to the Controller.
        """

        super().__init__(control_server)

        # Create a new Controller for Process Management

        self.controller = ProcessManagerController()

        # Load the commands (for commanding of the PM Controller) from the
        # YAML file into a dictionary, stored in the PM Protocol

        self.load_commands(
            COMMAND_SETTINGS.Commands, ProcessManagerCommand, ProcessManagerController
        )

        # Build a look-up table for the methods

        self.build_device_method_lookup_table(self.controller)

    def get_bind_address(self):
        """Returns the address to bind a socket to.

        This bind address is a properly formatted URL, based on the
        communication protocol and the commanding port.

        Returns:
            - Properly formatted URL to bind a socket to.
        """

        return bind_address(
            self.control_server.get_communication_protocol(),
            self.control_server.get_commanding_port(),
        )

    def get_status(self) -> dict:
        """Returns the status information for the Control Server.

        This status information is returned in the form of a dictionary that
        contains the following information about the Control Server for
        Process Management:

            - timestamp (str): string representation of the current datetime;
            - PID (int): process ID for the Control Server;
            - Up (float): uptime of the Control Server [s];
            - UUID (uuid1): Universally Unique Identifier for the Control
                            Server;
            - RSS (int): 'Resident Set Size', this is the non-swapped physical
                         memory a process has used [byte];
            - USS (int): 'Unique Set Size', this is the memory which is unique
                         to a process [byte];
            - CPU User (float): time spent in user mode [s];
            - CPU System (float): time spent in kernel mode [s];
            - CPU count: number of CPU cores in use by the process;
            - CPU% (float): process CPU utilization as a percentage [%].

        Returns:
            - Dictionary with status information for the Control Server for
              Process Management.
        """

        return super().get_status()

    def get_housekeeping(self) -> dict:
        """Returns the housekeeping data for the Control Server.

        This housekeeping data is returns in the form of a dictionary that
        contains the following information about the Control Server for
        Process Management:

            - timestamp (str): string representation of the current datetime.

        Returns:
            - Dictionary with housekeeping data for the Control Server for
              Process Management.
        """

        return {"timestamp": format_datetime()}

    def quit(self):
        self.controller.quit()


class ProcessManagerCommand(ClientServerCommand):

    """
    Command (client-server) for Process Management.
    """

    pass


class ProcessManagerInterface:

    """
    Interface for dynamic loading of the commands for Process Management.
    """

    @dynamic_interface
    def get_cm_proxy(self):
        """
        Returns the Proxy for Configuration Management.

        Returns:
            - Proxy for Configuration Management.
        """

        raise NotImplementedError

    @dynamic_interface
    def start_cs(self, process_name, sim_mode):
        """Start the device Control Server with the given process name.

        This method can only be used to start the Control Server of a device
        that is included in the current setup.  Core processes (Storage,
        Configuration Manager, and Process Manager) cannot be started with
        this method.

        The Control Server can either be started in simulator mode (without
        H/W controller available) or in operational mode (with H/W controller
        available).

        The given process name is the one that is used as device name in the
        setup file and can also be found in the dictionary returned by the
        get_devices() method.

        Args:
            - process_name: Device name for which the Control Server should be
                            started.
            - sim_mode: Whether or not to start the Control Server in
                        simulator mode.
        """

        raise NotImplementedError

    @dynamic_interface
    def start_egse(self):
        """Start all device Control Servers in the current setup in operational mode.

        This method can only be used to start the Control Servers of the
        devices that are included in the current setup.  Core processes
        (Storage, Configuration Manager, and Process Manager) cannot be
        started with this method.
        """

        raise NotImplementedError

    @dynamic_interface
    def shut_down_egse(self, process_name):
        """Shut down the device Control Server with the given process name.

        This method can only be used to shut down the Control Server of a
        device that is included in the current setup.  Core processes (Storage,
        Configuration Manager, and Process Manager) cannot be shut down with
        this method.

        The given process name is the one that is used as device name in the
        setup file and can also be found in the dictionary returned by the
        get_devices() method.

        Args:
            - process_name: Device name for which the Control Server should be
                            shut down.
        """

    @dynamic_interface
    def shut_down_egse(self):
        """Shut down all device Control Servers in the current setup.

        This method can only be used to shut down the Control Servers of the
        devices that are included in the current setup.  Core processes
        (Storage, Configuration Manager, and Process Manager) cannot be shut
        down with this method.
        """

        raise NotImplementedError

    @dynamic_interface
    def start_fitsgen(self):
        """ Start the FITS generation."""

        raise NotImplementedError

    @dynamic_interface
    def stop_fitsgen(self):
        """ Stop the FITS generation."""

        raise NotImplementedError

    @dynamic_interface
    def start_fov_hk(self):
        """ Start the generation of FOV HK."""

        raise NotImplementedError

    @dynamic_interface
    def stop_fov_hk(self):
        """ Stop the generation of FOV HK."""

        raise NotImplementedError

    @dynamic_interface
    def start_n_fee_hk(self):
        """ Start the generation of N-FEE HK."""

        raise NotImplementedError

    @dynamic_interface
    def stop_n_fee_hk(self):
        """ Stop the generation of N-FEE HK."""

        raise NotImplementedError


class ProcessManagerController(ProcessManagerInterface):

    """
    Controller for Process Management.
    """

    def __init__(self):
        """Initialisation for the Process Manager Controller.

        Bother the Configuration Manager and the Storage manager should be running.
        """

        # Keep track of the Configuration Manager

        self._configuration = ConfigurationManagerProxy()

        # Storage Manager must be active

        if not is_storage_manager_active():

            LOGGER.error("No Storage Manager available!!!!")

    def quit(self):

        self._configuration.disconnect_cs()

    def get_cm_proxy(self):
        """
        Returns the Proxy for Configuration Management.

        Returns:
            - Proxy for Configuration Management.
        """

        if not self._configuration.has_commands():
            self._configuration.load_commands()

        return self._configuration

    def get_devices(self):
        """Returns a dictionary with the device processes.

        The devices processes that are listed in this dictionary are the ones
        that are included in the current setup.

        The keys in the dictionary are taken from the "device_name" entries in
        the setup file. The corresponding values in the dictionary are taken
        from the "device" entries in the setup file (and should be Proxy classes).

        Returns:
            - Dictionary with the devices that are included in the setup.
        """

        try:

            setup = GlobalState.setup

            devices = {}
            devices = Setup.find_devices(setup, devices=devices)

            return devices

        except AttributeError:

            return {}

    def get_core(self):
        """Returns a dictionary with the core EGSE processes.

        The core EGSE processes are:

            - the Storage Manager,
            - the Configuration Manager,
            - the Process Manager,
            - and the Synoptics Manager.

        These processes should be running at all times, and can neither be
        started nor shut down from within the Process Manager.

        The keys in the dictionary are the names of the core processes.  The
        values are the Proxy classes.

        Returns:
            - Dictionary with the core EGSE processes.

        """

        storage_proxy_class = "class//egse.storage.StorageProxy"
        confman_proxy_class = "class//egse.confman.ConfigurationManagerProxy"
        procman_proxy_class = "class//egse.procman.ProcessManagerProxy"
        syn_proxy_class = "class//egse.synoptics.SynopticsManagerProxy"

        return {
            "Storage": (storage_proxy_class, ()),
            "Configuration Manager": (confman_proxy_class, ()),
            "Process Manager": (procman_proxy_class, ()),
            "Synoptics Manager": (syn_proxy_class, ()),
        }

    def start_egse(self):
        """Start all device Control Servers in the current setup in operational mode.

        This method can only be used to start the Control Servers of the
        devices that are included in the current setup.  Core processes
        (Storage, Configuration Manager, and Process Manager) cannot be
        started with this method.
        """

        LOGGER.debug("Starting EGSE")

        devices = self.get_devices()

        for process_name, process_info in devices.items():

            proxy_type = process_info[0]
            device_args = process_info[1]

            try:

                with find_class(proxy_type)(*device_args):

                    # The CS is already running

                    LOGGER.info(f"{process_name} was already running")

            except ConnectionError:

                try:

                    module_name = proxy_type[7:].rsplit(".", 1)[0]
                    module = importlib.import_module(module_name)

                    cs_type = module.DEVICE_SETTINGS.ControlServer

                    # Operational mode

                    if str.startswith(proxy_type, "class//egse.aeu.aeu.CRIO"):

                        cs = SubProcess("MyApp", [sys.executable, "-m", cs_type, "start-crio-cs"])

                    elif str.startswith(proxy_type, "class//egse.aeu.aeu.PSU"):

                        cs = SubProcess("MyApp", [sys.executable, "-m", cs_type, "start-psu-cs", str(device_args[0])])

                    elif str.startswith(proxy_type, "class//egse.aeu.aeu.AWG"):

                        cs = SubProcess("MyApp", [sys.executable, "-m", cs_type, "start-awg-cs", str(device_args[0])])

                    else:
                        if len(device_args) == 0:
                            cs = SubProcess("MyApp", [sys.executable, "-m", cs_type, "start"])
                        else:
                            cs = SubProcess("MyApp", [sys.executable, "-m", cs_type, "start", str(device_args[0])])

                    LOGGER.info(f"Starting Control Server for {process_name} in operational mode")
                    cs.execute(detach_from_parent=True)

                    if process_name == "DAQ6510":
                        setup = GlobalState.setup
                        das_delay = setup.gse.DAQ6510.route.mon_delay
                        das_count = setup.gse.DAQ6510.route.scan.COUNT.SCAN
                        das_interval = setup.gse.DAQ6510.route.scan.INTERVAL

                        das = SubProcess("MyApp", [sys.executable, "-m", "egse.das", "daq6510", "--count",
                                                   str(das_count), "--interval", str(das_interval), "--delay",
                                                   str(das_delay)])
                        das.execute(detach_from_parent="True")

                    # Check to see the CS actually started

                    time.sleep(5)
                    try:
                        with find_class(proxy_type)(*device_args):
                            pass
                    except ConnectionError:
                        LOGGER.warning(f"Could not start Control Server for {process_name}")

                except AttributeError:

                    LOGGER.debug(f"Cannot start Control Server for {process_name}")

        LOGGER.debug("EGSE started")

    def start_cs(self, process_name, sim_mode):
        """Start the device Control Server with the given process name.

        This method can only be used to start the Control Server of a device
        that is included in the current setup.  Core processes (Storage,
        Configuration Manager, and Process Manager) cannot be started with
        this method.

        The Control Server can either be started in simulator mode (without
        H/W controller available) or in operational mode (with H/W controller
        available).

        The given process name is the one that is used as device name in the
        setup file and can also be found in the dictionary returned by the
        get_devices() method.

        Args:
            - process_name: Device name for which the Control Server should be
                            started.
            - sim_mode: Whether or not to start the Control Server in
                        simulator mode.
        """

        LOGGER.debug(f"Starting {process_name}")

        process_info = self.get_devices()[process_name]
        proxy_type = process_info[0]
        device_args = process_info[1]

        try:

            with find_class(proxy_type)(*device_args):

                # The CS is already running

                message = f"{process_name} was already running"
                LOGGER.info(message)
                return Success(message)

        except ConnectionError:

            try:

                module_name = proxy_type[7:].rsplit(".", 1)[0]
                module = importlib.import_module(module_name)

                if hasattr(module, "DEVICE_SETTINGS") and hasattr(module.DEVICE_SETTINGS, "ControlServer"):
                    cs_type = module.DEVICE_SETTINGS.ControlServer
                else:
                    raise AttributeError(f"DEVICE_SETTINGS (or ControlServer therein) not defined in {module_name}")

                # Simulator mode

                if sim_mode:

                    mode = "simulator mode"

                    if str.startswith(proxy_type, "class//egse.aeu.aeu.CRIO"):

                        cs = SubProcess("MyApp", [sys.executable, "-m", cs_type, "start-crio-cs", "--sim"])

                    elif str.startswith(proxy_type, "class//egse.aeu.aeu.PSU"):

                        cs = SubProcess("MyApp", [sys.executable, "-m", cs_type, "start-psu-cs", str(device_args[0]), "--sim"])

                    elif str.startswith(proxy_type, "class//egse.aeu.aeu.AWG"):

                        cs = SubProcess("MyApp", [sys.executable, "-m", cs_type, "start-awg-cs", str(device_args[0]), "--sim"])

                    else:

                        cs = SubProcess("MyApp", [sys.executable, "-m", cs_type, "start", "--sim"])

                # Operational mode

                else:

                    mode = "operational mode"

                    if str.startswith(proxy_type, "class//egse.aeu.aeu.CRIO"):

                        cs = SubProcess("MyApp", [sys.executable, "-m", cs_type, "start-crio-cs"])

                    elif str.startswith(proxy_type, "class//egse.aeu.aeu.PSU"):

                        cs = SubProcess("MyApp", [sys.executable, "-m", cs_type, "start-psu-cs", str(device_args[0])])

                    elif str.startswith(proxy_type, "class//egse.aeu.aeu.AWG"):

                        cs = SubProcess("MyApp", [sys.executable, "-m", cs_type, "start-awg-cs", str(device_args[0])])

                    else:
                        if len(device_args) == 0:
                            cs = SubProcess("MyApp", [sys.executable, "-m", cs_type, "start"])
                        else:
                            cs = SubProcess("MyApp", [sys.executable, "-m", cs_type, "start", str(device_args[0])])
                    # os.system(sys.executable + " -m " + cs_type + " start")

                LOGGER.info(f"Starting Control Server for {process_name} in {mode}")
                cs.execute(detach_from_parent=True)

                if process_name == "DAQ6510":
                    das = SubProcess("MyApp", [sys.executable, "-m", "egse.das", "daq6510"])
                    das.execute(detach_from_parent="True")

                # Check to see the CS actually started

                time.sleep(5)
                try:
                    with find_class(proxy_type)(*device_args):
                        return Success(f"{process_name} successfully started")
                except ConnectionError as exc:
                    message = f"Could not start Control Server for {process_name}"
                    LOGGER.warning(message, exc_info=True)
                    return Failure(message, cause=exc)

            except AttributeError as exc:
                message = f"Cannot start Control Server for {process_name}"
                LOGGER.error(message, exc_info=True)
                return Failure(message, cause=exc)

        # with find_class(proxy_type)(*device_args):
        #         pass

    def shut_down_egse(self):
        """Shut down all device Control Servers in the current setup.

        This method can only be used to shut down the Control Servers of the
        devices that are included in the current setup.  Core processes
        (Storage, Configuration Manager, and Process Manager) cannot be shut
        down with this method.
        """

        LOGGER.debug("Shutting down EGSE")

        devices = self.get_devices()

        for key, process_info in reversed(devices.items()):

            LOGGER.debug(f"Shutting down {key}")

            proxy_type = process_info[0]
            device_args = process_info[1]

            try:

                with find_class(proxy_type)(*device_args) as process_proxy:

                    with process_proxy.get_service_proxy() as service_proxy:

                        service_proxy.quit_server()

            except ConnectionError:

                pass

        LOGGER.debug("EGSE shut down")

    def shut_down_cs(self, process_name):
        """Shut down the device Control Server with the given process name.

        This method can only be used to shut down the Control Server of a
        device that is included in the current setup.  Core processes (Storage,
        Configuration Manager, and Process Manager) cannot be shut down with
        this method.

        The given process name is the one that is used as device name in the
        setup file and can also be found in the dictionary returned by the
        get_devices() method.

        Args:
            - process_name: Device name for which the Control Server should be
                            shut down.
        """

        LOGGER.debug(f"Shutting down {process_name}")

        try:

            process_info = self.get_devices()[process_name]
            proxy_type = process_info[0]
            device_args = process_info[1]

            with find_class(proxy_type)(*device_args) as process_proxy:

                with process_proxy.get_service_proxy() as service_proxy:

                    LOGGER.debug("Shutting down CS")

                    service_proxy.quit_server()

        except ConnectionError:

            # The CS is already down

            LOGGER.info(f"{process_name} was already down")

    def start_fitsgen(self):
        """ Start the FITS generation."""

        # TODO Think about potential conditions for the FITS generator to be started:
        #   - FITS generator should not be running yet
        #   - FEE simulator must be running?
        #   - DPU CS must be running?

        # Check whether the DPU CS is running

        if not is_dpu_cs_active():
            message = "The DPU Control Server must be running to be able to start the FITS generation."
            LOGGER.critical(message)
            return Failure(message)

        LOGGER.info("Starting the FITS generation")

        fg = SubProcess("MyApp", [sys.executable, "-m", "egse.dpu.fitsgen", "start"])
        fg.execute(detach_from_parent=True)

        time.sleep(5)
        if fitsgen.send_request("status").get("status") != "ACK":
            return Failure("FITS generation could not be started for some unknown reason.")

        return Success("FITS generation successfully started.")

    def stop_fitsgen(self):
        """ Stop the FITS generation."""

        LOGGER.info("Stopping the FITS generation")

        fg = SubProcess("MyApp", [sys.executable, "-m", "egse.dpu.fitsgen", "stop"])
        fg.execute(detach_from_parent=True)

    def start_fov_hk(self):
        """ Start the generation of FOV HK."""

        LOGGER.info("Starting the generation of FOV HK")

        fg = SubProcess("MyApp", [sys.executable, "-m", "egse.fov.fov_hk", "start"])
        fg.execute(detach_from_parent=True)

        time.sleep(5)
        if fov_hk.send_request("status").get("status") != "ACK":
            return Failure("FOV HK generation could not be started for some unknown reason.")

        return Success("FOV HK generation successfully started.")

    def stop_fov_hk(self):
        """ Stop the generation of FOV HK."""

        LOGGER.info("Stopping the generation of FOV HK")

        fg = SubProcess("MyApp", [sys.executable, "-m", "egse.fov.fov_hk", "stop"])
        fg.execute(detach_from_parent=True)

    def start_n_fee_hk(self):
        """ Start the generation of N-FEE HK."""

        LOGGER.info("Starting the generation of N-FEE HK")

        fg = SubProcess("MyApp", [sys.executable, "-m", "egse.fee.n_fee_hk", "start"])
        fg.execute(detach_from_parent=True)

        time.sleep(5)
        if n_fee_hk.send_request("status").get("status") != "ACK":
            return Failure("N-FEE HK generation could not be started for some unknown reason.")

        return Success("N-FEE HK generation successfully started.")

    def stop_n_fee_hk(self):
        """ Stop the generation of N-FEE HK."""

        LOGGER.info("Stopping the generation of N-FEE HK")

        fg = SubProcess("MyApp", [sys.executable, "-m", "egse.fee.n_fee_hk", "stop"])
        fg.execute(detach_from_parent=True)


class ProcessManagerProxy(Proxy, ProcessManagerInterface):

    """
    Proxy for Process Managements, used to connect to the Process Manager
    Control Server and send commands remotely.
    """

    def __init__(
        self,
        protocol=CTRL_SETTINGS.PROTOCOL,
        hostname=CTRL_SETTINGS.HOSTNAME,
        port=CTRL_SETTINGS.COMMANDING_PORT,
    ):
        """Initialisation of a new Proxy for Process Management.

        If no connection details (transport protocol, hostname, and port) are
        not provided, these are taken from the settings file.

        Args:
            - protocol: Transport protocol [default is taken from settings
                        file].
            - hostname: Location of the control server (IP address) [default
                        is taken from settings file].
            - port: TCP port on which the Control Server is listening for
                    commands [default is taken from settings file].
        """

        super().__init__(connect_address(protocol, hostname, port))



"""
This process doesn't do anything and serves as a skeleton for testing sub-processes.

This process listens on port 5556 for ZeroMQ messages (using the REQ-REP protocol)
and understands the following commands:

* Ping: returns "Pong" as an answer
* Quit: returns "Quiting" as an answer
* Status?: returns status information of the SubProcess

"""

import logging
import pickle
import sys
import time

import zmq

from egse.process import ProcessStatus

LOGGER = logging.getLogger("egse.tests")

PORT = 5556


def main():
    status = ProcessStatus()

    context = zmq.Context.instance()

    process_socket: zmq.Socket = context.socket(zmq.REP)
    process_socket.bind(f"tcp://*:{PORT}")

    while True:
        #  Wait for next request from client
        try:
            message = pickle.loads(process_socket.recv())
        except KeyboardInterrupt:
            LOGGER.info("Keyboard interrupt, terminating...")
            break

        LOGGER.info(f"Received request: {message}")

        #  Send reply back to client

        if message == "Ping":
            process_socket.send(pickle.dumps("Pong"))
        if message == "Quit":
            process_socket.send(pickle.dumps("Quiting"))
            break
        if message == "Status?":
            status.update()
            process_socket.send(pickle.dumps(str(status)))

        #  Do some 'work'
        time.sleep(0.1)

    process_socket.close(linger=0)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    sys.exit(main())



from __future__ import annotations

import asyncio
import json
import logging
import time
from typing import Any
from typing import Callable

import zmq
import zmq.asyncio

from egse.registry import DEFAULT_RS_PUB_PORT
from egse.registry import DEFAULT_RS_REQ_PORT
from egse.registry.client import AsyncRegistryClient
from egse.system import get_host_ip
from egse.zmq_ser import get_port_number

module_module_logger_name = "async_microservice"
module_logger = logging.getLogger(module_module_logger_name)


class ZMQMicroservice:
    """
    A pure ZeroMQ-based microservice implementation.

    This service:
    1. Registers with the service registry.
       Default endpoints are "tcp://localhost:4242" for the REQ-REP, and
       "tcp://localhost:4243" for the PUB-SUB (events, notifications)
    2. Exposes ZeroMQ sockets for its functionality
    3. Can discover and call other services
    """

    def __init__(
        self,
        service_name: str,
        service_type: str,
        rep_port: int = 0,
        pub_port: int | None = None,
        registry_req_endpoint: str | None = None,
        registry_sub_endpoint: str | None = None,
        metadata: dict[str, Any] | None = None
    ):
        """
        Initialize the microservice. When port numbers for rep_port and pub_port
        are set to 0, the system will automatically assign a dynamic port number
        that will be reported to the service registry.

        Args:
            service_name: Human-readable name for this service
            service_type: Type of service (for discovery)
            rep_port: Port for REP socket (service API) [default=0]
            pub_port: Optional port for PUB socket (events/notifications)
            registry_req_endpoint: ZeroMQ endpoint for registry REQ socket
            registry_sub_endpoint: ZeroMQ endpoint for registry SUB socket
            metadata: Additional service metadata
        """
        self.service_name = service_name
        self.service_type = service_type
        self.rep_port = rep_port
        self.pub_port = pub_port
        self.registry_req_endpoint = registry_req_endpoint or f"tcp://localhost:{DEFAULT_RS_REQ_PORT}"
        self.registry_sub_endpoint = registry_sub_endpoint or f"tcp://localhost:{DEFAULT_RS_PUB_PORT}"
        self.metadata = metadata or {}

        self.host_ip = get_host_ip()

        # Service ID will be set when registered
        self.service_id = None

        # ZeroMQ context
        self.context = zmq.asyncio.Context()

        # REP socket for service API
        self.rep_socket = self.context.socket(zmq.REP)
        self.rep_socket.bind(f"tcp://*:{rep_port}")

        # Determine the dynamically assigned port number
        if rep_port == 0:
            self.rep_port = get_port_number(self.rep_socket)

        # PUB socket for events/notifications (optional)
        if pub_port is not None:
            self.pub_socket = self.context.socket(zmq.PUB)
            self.pub_socket.bind(f"tcp://*:{pub_port}")
        else:
            self.pub_socket = None

        # Determine the dynamically assigned port number
        if pub_port == 0:
            self.pub_port = get_port_number(self.pub_socket)

        self.registry_client = AsyncRegistryClient(
            registry_req_endpoint=self.registry_req_endpoint,
            registry_sub_endpoint=self.registry_sub_endpoint
        )

        self.command_handlers = {}

        self._register_default_handlers()

        # Shutdown will be 'set' to terminate the service.
        self._shutdown = asyncio.Event()

        self._tasks = set()

    def _register_default_handlers(self):
        """Register default command handlers."""
        self.register_handler("ping", self._handle_ping)
        self.register_handler("info", self._handle_info)
        self.register_handler("health", self._handle_health)

    def register_handler(self, command: str, handler: Callable):
        """
        Register a command handler.

        Args:
            command: Command name
            handler: Async function that will handle the command
        """
        self.command_handlers[command] = handler
        module_logger.info(f"Registered handler for command: {command}")

    async def _handle_ping(self, data: dict[str, Any]) -> dict[str, Any]:
        """Handle ping command."""
        return {
            "status": "ok",
            "message": "pong",
            "timestamp": time.time()
        }

    async def _handle_info(self, data: dict[str, Any]) -> dict[str, Any]:
        """Handle info command."""
        return {
            "status": "ok",
            "service_id": self.service_id,
            "service_name": self.service_name,
            "service_type": self.service_type,
            "host": self.host_ip,
            "rep_port": self.rep_port,
            "pub_port": self.pub_port,
            "metadata": self.metadata
        }

    async def _handle_health(self, data: dict[str, Any]) -> dict[str, Any]:
        """Handle health command."""
        return {
            "status": "ok",
            "timestamp": time.time()
        }

    async def start(self):
        """Start the microservice. This will:

        - register the service to the service registry
        - start the service heartbeat
        - start the service event listener
        - start the background task to handle requests
        """
        module_logger.info(f"Starting {self.service_name} ({self.service_type}) on {self.host_ip}:{self.rep_port}")

        # Register with the registry
        self.service_id = await self.registry_client.register(
            self.service_name,
            self.host_ip,
            self.rep_port,
            service_type=self.service_type,
            metadata={
                **self.metadata,
                "pub_port": self.pub_port
            }
        )

        if not self.service_id:
            module_logger.error("Failed to register with the service registry")
            return True

        module_logger.info(f"Registered with service ID: {self.service_id}")

        # Start heartbeat and event listener
        await self.registry_client.start_heartbeat()
        await self.registry_client.start_event_listener()

        # Start request handler
        request_task = asyncio.create_task(self._handle_requests())
        self._tasks.add(request_task)
        request_task.add_done_callback(self._tasks.discard)

        # Wait for shutdown signal
        await self._shutdown.wait()

        # Clean shutdown
        await self._cleanup()

        return False

    async def _handle_requests(self):
        """Handle incoming requests."""
        module_logger.info("Started request handler")

        try:
            while not self._shutdown.is_set():
                try:
                    # Wait for a request with timeout
                    try:
                        request_json = await asyncio.wait_for(
                            self.rep_socket.recv_string(),
                            timeout=1.0
                        )
                    except asyncio.TimeoutError:
                        continue

                    # Parse the request
                    request = json.loads(request_json)
                    command = request.get("command")

                    if not command:
                        response = {
                            "status": "error",
                            "error": "Missing 'command' field"
                        }
                    elif command not in self.command_handlers:
                        response = {
                            "status": "error",
                            "error": f"Unknown command: {command}"
                        }
                    else:
                        # Call the handler
                        try:
                            handler = self.command_handlers[command]
                            response = await handler(request)
                        except Exception as exc:
                            module_logger.error(f"Error handling command {command}: {exc}")
                            response = {
                                "status": "error",
                                "error": f"Error handling command: {str(exc)}"
                            }

                    # Send the response
                    await self.rep_socket.send_string(json.dumps(response))
                except zmq.ZMQError as exc:
                    module_logger.error(f"ZMQ error: {exc}")
                except json.JSONDecodeError:
                    module_logger.error("Invalid JSON received")
                    try:
                        await self.rep_socket.send_string(json.dumps(
                            {
                                "status": "error",
                                "error": "Invalid JSON format"
                            }
                        ))
                    except Exception as exc:
                        module_logger.warning(f"Caught {type(exc).__name__}: {exc}")
                except Exception as exc:
                    module_logger.error(f"Error handling request: {exc}")
                    try:
                        await self.rep_socket.send_string(json.dumps(
                            {
                                "status": "error",
                                "error": str(exc)
                            }
                        ))
                    except Exception as exc:
                        module_logger.warning(f"Caught {type(exc).__name__}: {exc}")
        except asyncio.CancelledError:
            module_logger.info("Request handler task cancelled")

    async def stop(self):
        """Signal the service to stop."""
        self._shutdown.set()

    async def _cleanup(self):
        """Clean up resources."""
        module_logger.info("Cleaning up resources...")

        # Cancel all tasks
        for task in self._tasks:
            task.cancel()

        if self._tasks:
            await asyncio.gather(*self._tasks, return_exceptions=True)

        # Deregister from the registry
        if self.service_id:
            await self.registry_client.deregister()
            module_logger.info(f"Deregistered service: {self.service_id}")

        # Close registry client
        await self.registry_client.close()

        # Close ZeroMQ sockets
        if hasattr(self, 'rep_socket') and self.rep_socket:
            self.rep_socket.close()

        if hasattr(self, 'pub_socket') and self.pub_socket:
            self.pub_socket.close()

        if hasattr(self, 'context') and self.context:
            self.context.term()

        module_logger.info("Cleanup complete")

    async def publish_event(self, event_type: str, data: dict[str, Any]):
        """
        Publish an event to subscribers.

        Args:
            event_type: Type of event
            data: Event data
        """
        if not self.pub_socket:
            module_logger.warning("Cannot publish event: no PUB socket")
            return

        event = {
            "type": event_type,
            "service_id": self.service_id,
            "timestamp": time.time(),
            "data": data
        }

        try:
            await self.pub_socket.send_multipart([
                event_type.encode(
                    'utf-8'),
                    json.dumps(event).encode('utf-8')
            ])
            module_logger.debug(f"Published event: {event_type}")
        except Exception as exc:
            module_logger.error(f"Failed to publish event: {exc}")

    async def call_service(self, service_type: str, command: str, data: dict[str, Any] | None = None) -> dict[str, Any]:
        """
        Call another service.

        Args:
            service_type: Type of service to call
            command: Command to invoke
            data: Command parameters

        Returns:
            Response from the service
        """
        # Discover the service
        service = await self.registry_client.discover_service(service_type)

        if not service:
            raise ValueError(f"No service of type '{service_type}' found")

        # Prepare the request
        request = {
            "command": command,
            **(data or {})
        }

        # Create a REQ socket
        socket = self.context.socket(zmq.REQ)
        socket.setsockopt(zmq.RCVTIMEO, 5000)  # 5 second timeout

        try:
            # Connect to the service
            socket.connect(f"tcp://{service['host']}:{service['port']}")

            # Send the request
            await socket.send_string(json.dumps(request))

            # Wait for the response
            response_json = await socket.recv_string()
            return json.loads(response_json)
        finally:
            socket.close()

    async def subscribe_to_service_events(
            self, service_type: str, event_types: list[str], callback: Callable[[dict[str, Any]], None]):
        """
        Subscribe to events from another service.

        Args:
            service_type: Type of service to subscribe to
            event_types: List of event types to subscribe to
            callback: Function to call with events

        Returns:
            Subscription task
        """
        # Discover the service
        service = await self.registry_client.discover_service(service_type)

        if not service:
            raise ValueError(f"No service of type '{service_type}' found")

        # Get PUB port from metadata
        pub_port = service.get('metadata', {}).get('pub_port')

        if not pub_port:
            raise ValueError("Service does not expose a PUB socket")

        # Create a SUB socket
        socket = self.context.socket(zmq.SUB)

        # Subscribe to specified event types
        for event_type in event_types:
            socket.setsockopt_string(zmq.SUBSCRIBE, event_type)

        # Connect to the service
        socket.connect(f"tcp://{service['host']}:{pub_port}")

        async def subscription_loop():
            try:
                module_logger.info(f"Subscribed to events from {service['name']} ({service_type})")

                while not self._shutdown.is_set():
                    try:
                        # Wait for an event with timeout
                        if socket.poll(timeout=1000) == 0:
                            continue

                        # Receive the event
                        event_type_bytes, event_json_bytes = await socket.recv_multipart()
                        event_type = event_type_bytes.decode('utf-8')
                        event = json.loads(event_json_bytes.decode('utf-8'))

                        # Call the callback
                        try:
                            if asyncio.iscoroutinefunction(callback):
                                await callback(event)
                            else:
                                callback(event)
                        except Exception as exc:
                            module_logger.error(f"Error in event callback: {exc}")
                    except zmq.ZMQError as exc:
                        if exc.errno == zmq.EAGAIN:
                            # Timeout, just continue
                            continue
                        module_logger.error(f"ZMQ error in subscription: {exc}")
                    except Exception as exc:
                        module_logger.error(f"Error in subscription loop: {exc}")
                        await asyncio.sleep(1)  # Prevent tight loop on error
            except asyncio.CancelledError:
                module_logger.info(f"Subscription to {service_type} cancelled")
            finally:
                socket.close()

        # Start the subscription task
        task = asyncio.create_task(subscription_loop())
        self._tasks.add(task)
        task.add_done_callback(self._tasks.discard)

        return task



"""
Test Mixin functions that are defined in the egse.mixin module.
"""

import logging
import random
from enum import IntEnum
from typing import Callable

import pytest

from egse.command import CommandError
from egse.decorators import dynamic_interface
from egse.decorators import read_command
from egse.device import DeviceTransport
from egse.mixin import DynamicCommandMixin
from egse.mixin import dynamic_command

LOGGER = logging.getLogger("egse.test.mixin")

CHOICES = [b"I don't know what you mean.", b"Can you repeat the question, please?", b"Are you talking to me?"]


# The following two functions shall move to egse.decorators


def query_command(cmd_string: str = None):
    """Decorate an interface function for querying."""

    def func_wrapper(func):
        """Adds a static variable `__query_command` to a method."""
        setattr(func, "__query_command", True)
        setattr(func, "__cmd_string", cmd_string)
        return func

    return func_wrapper


def write_command(cmd_string: str = None):
    """Decorate an interface function for writing."""

    def func_wrapper(func):
        """Adds a static variable `__query_command` to a method."""
        setattr(func, "__write_command", True)
        setattr(func, "__cmd_string", cmd_string)
        return func

    return func_wrapper


def response_processor(processor: Callable):
    """Decorate an interface with a function to process the response of the instrument command."""

    def func_wrapper(func):
        """Adds a static variable `__process_response` to a method."""
        setattr(func, "__process_response", processor)
        return func

    return func_wrapper


def reverse_response(response: bytes) -> bytes:
    """Reverse a bytes object."""
    return response[::-1]


def index_of(response: bytes) -> int:
    """Return the index in CHOICES."""
    return CHOICES.index(response)


class FriendlyTransport(DeviceTransport):
    """What goes to the device and comes back as response."""

    def write(self, command: str):
        LOGGER.info(f"I'm writing down: {command}")

    def read(self) -> bytes:
        return random.choice(CHOICES)

    def trans(self, command: str) -> bool:
        LOGGER.info(f"I'm answering: {command}")
        return False if "studying" in command else True


class FriendlyInterface:
    """Interface definition of all valid commands."""

    @dynamic_interface
    @read_command
    def how_do_you_do(self) -> str:
        """The easiest question."""

    @dynamic_interface
    @query_command("Can I help you with ${topic}?")
    def can_i_help_you_with(self, topic: str) -> bool:
        """Always returns True, except for topic='studying'."""

    @dynamic_interface
    @write_command("Can you do ${command} for me, please?")
    def can_you_do(self, command: str) -> bool:
        """Print an almost friendly welcome message."""

    @dynamic_interface
    @write_command("Can we do ${command} with ${who}?")
    def can_we_do(self, command: str, *, who: str) -> bool:
        """Print an almost friendly welcome message."""

    @dynamic_interface
    @read_command
    @response_processor(reverse_response)
    def can_you_hear_me(self):
        """Return a reverse of the response."""

    @dynamic_interface
    @read_command
    @response_processor(index_of)
    def which_choice(self):
        """Return the choice that was made."""


class Questions(DynamicCommandMixin, FriendlyInterface):
    """Ask some friendly questions."""

    def __init__(self):
        self.transport = FriendlyTransport()
        super().__init__()

    def are_you_directed(self) -> str:
        """Answers directly."""
        return "No, I'm not directed."


def test_standard_commands():
    q = Questions()

    assert q.how_do_you_do() in CHOICES  # read_command
    assert "not directed" in q.are_you_directed()  # command defined in Questions class itself
    assert q.can_you_do(command="the laundry") is None  # write_command
    assert not q.can_i_help_you_with(topic="studying")  # query_command
    assert q.can_i_help_you_with(topic="the laundry")  # query_command

    with pytest.raises(CommandError):
        q.can_we_do("a BBQ", "friends")  # write_command

    q.can_we_do("a BBQ", who="friends")  # write_command

    assert not q.can_i_help_you_with("studying")  # query_command


def test_docstring_example():
    """This example is given in the docstring of `create_command_string`."""

    def func(a, b, flag=True):
        pass

    template = "CREATE:FUN:${a} ${b} [${flag}]"

    response = DynamicCommandMixin.create_command_string(func, template, "TEMP", 23)
    assert response == "CREATE:FUN:TEMP 23 [True]"


def test_process_response():
    q = Questions()

    # maybe also test the reverse function first ;)

    assert reverse_response(b"abcdefgh") == b"hgfedcba"

    assert q.can_you_hear_me()[::-1] in CHOICES
    assert q.which_choice() in list(range(len(CHOICES)))


RESPONSES = [b"42", b"73", b"27", b"666", b"123"]


class NewStyleTransport(DeviceTransport):
    """What goes to the device and comes back as response."""

    def write(self, command: str):
        LOGGER.info(f"write: {command}")

    def read(self) -> bytes:
        return random.choice(RESPONSES)

    def trans(self, command: str) -> bytes:
        LOGGER.info(f"trans: {command}")
        return command.encode()

    def query(self, command: str) -> bytes:
        LOGGER.info(f"query: {command}")
        return random.choice(RESPONSES)


class MyEnum(IntEnum):
    XXX = 42
    YYY = 73


def pre_command(transport: DeviceTransport = None, **kwargs):
    LOGGER.info(f"Inside pre_command({transport.__class__.__name__}, {kwargs})")


def post_command(transport: DeviceTransport = None, response: bytes = None):
    LOGGER.info(f"Inside post_command(transport={transport.__class__.__name__}, {response=})")
    return response


class NewStyleCommandInterface:
    """Interface definition for new style dynamic commands."""

    @dynamic_command(cmd_type="read")
    def new_style_read_command(self) -> bool:
        """A read command only needs to set the cmd_type and doesn't take any arguments."""

    @dynamic_command(cmd_type="write", cmd_string="new style writing: flag=${flag}")
    def new_style_write_command(self, flag=True):
        """A write command shall also define cmd_string and usually has arguments."""

    @dynamic_command(cmd_type="query", cmd_string="new style query: request temp=${temp}")
    def new_style_query_command(self, temp: str):
        """A query command shall define both cmd_type and cmd_string."""

    @dynamic_command(cmd_type="transaction", cmd_string="new style transaction command: config=${config}")
    def new_style_trans_command(self, config: str = "CONFIG"):
        """A transaction command shall define both cmd_type and cmd_string."""

    @dynamic_command(
        cmd_type="transaction",
        cmd_string="trans_command, ${x}, ${y}, ${z}",
        process_cmd_string=lambda x: x.upper() + "\x0d\x0a",
    )
    def test_cmd_string_processor(self, x, y, *, z):
        """Returns the command string."""

    @dynamic_command(
        cmd_type="transaction", cmd_string="trans_command, ${x}, ${y}, ${z} and ${SOME_PREDEFINED_VARIABLE}"
    )
    def test_cmd_string_extra(self, x, y, *, z):
        """Returns the command string."""

    @dynamic_command(cmd_type="transaction", cmd_string="trans_command, {x:04d}, {y:0.2f}, {z:0.4f}", use_format=True)
    def test_cmd_string_format(self, x, y, *, z):
        """Returns the command string."""

    @dynamic_command(cmd_type="transaction", cmd_string="trans_command ${enumeration}")
    def test_cmd_enum(self, enumeration: MyEnum):
        """Returns the command string"""

    @dynamic_command(cmd_type="transaction", cmd_string="trans_command {enumeration}", use_format=True)
    def test_cmd_enum_with_format(self, enumeration: MyEnum):
        """Returns the command string"""

    @dynamic_command(cmd_type="transaction", cmd_string="msg = ${msg}", pre_cmd=pre_command, post_cmd=post_command)
    def test_cmd_pre_post(self, msg: str):
        """Return the command string, executes pre- and post-commands."""


class NewStyleCommand(DynamicCommandMixin, NewStyleCommandInterface):
    """Simple new style commanding interface."""

    def __init__(self):
        self.transport = NewStyleTransport()
        super().__init__()


def test_new_style(caplog):
    caplog.set_level(logging.INFO)

    print()

    ns = NewStyleCommand()

    # Test basic normal behaviour

    assert ns.new_style_read_command() in RESPONSES
    assert ns.new_style_write_command() is None
    assert ns.new_style_query_command("TRP22") in RESPONSES
    assert ns.new_style_trans_command() == b"new style transaction command: config=CONFIG"

    # Test some expected errors

    with pytest.raises(CommandError):
        ns.test_cmd_string_processor(2, 3, 36)

    # Test cmd_string processor

    assert ns.test_cmd_string_processor(2, 3, z=36) == b"TRANS_COMMAND, 2, 3, 36\r\n"

    # Check if the correct docstring is attached to the dynamic command

    assert "transaction" in ns.new_style_trans_command.__doc__

    # $-based substitutions that are not in the argument list will be left alone.
    # This allows to substitute these at another level, e.g. in the cmd_string processor, or
    # in the transport methods.

    assert ns.test_cmd_string_extra(1, 2, z=3) == b"trans_command, 1, 2, 3 and ${SOME_PREDEFINED_VARIABLE}"

    assert ns.test_cmd_string_extra(1, 2, z=None) == b"trans_command, 1, 2, None and ${SOME_PREDEFINED_VARIABLE}"

    assert ns.test_cmd_string_format(27, 42.12345, z=1.3) == b"trans_command, 0027, 42.12, 1.3000"

    assert ns.test_cmd_enum(MyEnum.XXX) == f"trans_command {MyEnum.XXX}".encode()

    assert ns.test_cmd_enum_with_format(MyEnum.XXX) == f"trans_command {MyEnum.XXX}".encode()

    caplog.clear()

    response = ns.test_cmd_pre_post("testing pre- and post-commands")
    assert response == b"msg = testing pre- and post-commands"

    print(f"{caplog.text = }")

    assert "Inside pre_command(NewStyleTransport" in caplog.text
    assert "function_name" in caplog.text
    assert (
        "Inside post_command(transport=NewStyleTransport, response=b'msg = testing pre- and post-commands')"
        in caplog.text
    )


def test_interface_definition():
    with pytest.raises(TypeError):

        class CommandInterface:
            """
            This will raise the following TypeError:
            dynamic_command() missing 1 required keyword-only argument: 'cmd_type'
            """

            @dynamic_command()
            def no_cmd_type(self):
                pass

    with pytest.raises(ValueError):

        class CommandInterface:
            """
            This will raise a ValueError: No cmd_string was provided for cmd_type='write'.
            """

            @dynamic_command(cmd_type="write")
            def no_cmd_string(self):
                pass

    with pytest.raises(ValueError):

        class CommandInterface:
            """
            This will raise a ValueError: No cmd_string was provided for cmd_type='query'.
            """

            @dynamic_command(cmd_type="query")
            def no_cmd_string(self):
                pass

    with pytest.raises(ValueError):

        class CommandInterface:
            """
            This will raise a ValueError: No cmd_string was provided for cmd_type='transaction'.
            """

            @dynamic_command(cmd_type="transaction")
            def no_cmd_string(self):
                pass



import logging
import os
import pickle
import sys
import time
import types
from pathlib import Path

import zmq

from egse.config import find_file
from egse.process import ProcessStatus
from egse.process import SubProcess
from egse.process import get_process_info
from egse.process import is_process_running
from egse.process import list_processes
from egse.process import list_zombies
from egse.process import ps_egrep
from egse.system import capture_rich_output

logger = logging.getLogger("egse.test.process")

HERE = Path(__file__).parent

# TODO
#   * How can we test detached processes?


def create_zombie():
    pid = os.fork()

    if pid == 0:  # Child process
        # Child immediately exits
        os._exit(0)  # noqa

    # Parent process continues

    print(f"Zombie child process created using 'fork' with PID: {pid}")

    return pid


def test_zombie():
    zombie_pid = create_zombie()

    zombies = list_zombies()
    print(f"{zombies = }")

    assert any([True for x in zombies if x["pid"] == zombie_pid])

    create_zombie()

    zombies = list_zombies()
    print(f"{zombies = }")

    assert len(list_zombies()) == 2


def test_is_process_running():
    # The empty_process.py should be located in the src/tests/scripts directory of the project

    stub = SubProcess("Stub Process", [sys.executable, str(find_file("empty_process.py").resolve())])
    stub.execute()

    time.sleep(0.1)  # allow process time to start/terminate

    list_processes(items=["python", "empty"])

    assert is_process_running(items=["python", "empty_process"])
    assert not is_process_running(items=["python", "Empty_process"], case_sensitive=True)  # command is lower case

    assert is_process_running(items=["python", "empty"], case_sensitive=True)

    assert is_process_running(items=["empty_process"])
    assert is_process_running(items="empty_process")

    assert not is_process_running(items=["_$123^_", "empty"])

    stub.quit()


def test_get_process_info():
    # The empty_process.py should be located in the src/tests/scripts directory of the project

    stub = SubProcess("Stub Process", [sys.executable, str(find_file("empty_process.py").resolve())])
    stub.execute()

    time.sleep(0.1)  # allow process time to start/terminate

    if len(get_process_info(["empty_process"])) > 1:
        logger.warning("Multiple process with 'empty_process' running.")

        # We would need this construct if multiple processes matching the criteria are running
        assert any(stub.pid == p["pid"] for p in get_process_info(["empty_process"]))
    else:
        assert get_process_info("empty_process")[0]["pid"] == stub.pid

    stub.quit()


def test_unknown_process():
    # The file unknown.exe does not exist and this will raise a FileNotFoundError which is caught in the execute()
    # method.
    process = SubProcess("Unknown App", ["unknown.exe"])

    assert not process.execute()
    time.sleep(0.5)  # allow process time to terminate
    logger.info(f"{process.exc_info = }")
    assert not process.is_running()
    assert process.returncode() is None


def test_error_during_execute():
    # The __file__  exists, but is not executable and will therefore raise a PermissionError which is caught in the
    # `execute()` method.

    process = SubProcess("Stub Process", [__file__])

    assert not process.execute()
    time.sleep(0.5)  # allow process time to terminate
    ei = process.exc_info
    print(f"{ei = }")
    assert ei["exc_type"] is PermissionError
    assert isinstance(ei["exc_value"], PermissionError)
    assert isinstance(ei["exc_traceback"], types.TracebackType)
    assert ei["command"] == __file__
    assert not process.is_running()
    assert process.returncode() is None


def test_terminated_process():
    # Process void-0 exits with an exit code of 0

    process = SubProcess("Stub Process", [sys.executable, str(find_file("void-0.py").resolve())])

    assert process.execute()
    time.sleep(0.5)  # allow process time to terminate
    assert not process.is_running()
    assert process.returncode() == 0

    # Process void-1 exits with an exit code of 1

    process = SubProcess("Stub Process", [sys.executable, str(find_file("void-1.py").resolve())])

    assert process.execute()
    time.sleep(0.5)  # allow process time to terminate
    assert not process.is_running()
    assert process.returncode() == 1


def test_quit_process():
    # when --ignore-sigterm is given, the process will be killed and return code will be -9.

    process = SubProcess(
        "Handle SIGTERM", [sys.executable, str(find_file("handle_sigterm.py").resolve()), "--ignore-sigterm"]
    )

    assert process.execute()
    time.sleep(1.0)  # allow process to start
    assert process.is_running()
    rc = process.quit()
    logger.info(f"After quit() -> {rc = }")
    assert rc == -9

    while process.is_running():
        logger.info(f"Process (PID={process.pid}) is still running...")
        time.sleep(1.0)

    assert process.returncode() == -9

    # when --ignore-sigterm is not given, the process will handle the SIGTERM and exit with 42.

    process = SubProcess("Handle SIGTERM", [sys.executable, str(find_file("handle_sigterm.py").resolve())])

    assert process.execute()
    time.sleep(1.0)  # allow process to start
    assert process.is_running()
    rc = process.quit()
    logger.info(f"After quit() -> {rc = }")
    assert rc == 42

    while process.is_running():
        logger.info(f"Process (PID={process.pid}) is still running...")
        time.sleep(1.0)

    assert process.returncode() == 0  # I would have expected 42 here, don't know why 0


def test_active_process():
    # The empty_process.py should be located in the tests/scripts directory of this project

    stub = SubProcess("Stub Process", [sys.executable, str(find_file("empty_process.py", root=HERE).resolve())])

    # We can set this cmd_port here because we know this from the empty_process.py file
    # In nominal situations, the cmd_port is known from the configuration file of the
    # system (because all processes are known) or communicated to the process manager
    # by the sub-process.

    cmd_port = 5556

    # Execute the sub-process.

    assert stub.execute()

    time.sleep(0.1)

    assert stub.is_running()
    assert is_active(cmd_port)  # check if the empty_process is active
    assert stub.returncode() is None

    status: str = get_status(cmd_port)
    logger.info(f"ProcessStatus: {status}")

    assert "PID" in status
    assert "UUID" in status
    assert "Up" in status

    time.sleep(1.0)

    logger.info(f"ProcessStatus: {get_status(cmd_port)}")

    procs = list_processes("empty_process", verbose=True)
    logger.info(f"{procs = }")

    assert quit_process(cmd_port)

    time.sleep(0.1)

    # if the sub-process is not in the processes list, it means it has terminated,
    # but since this is running under pytest, the sub-process will be a zombie process
    # until pytest finishes.

    procs = list_processes("empty_process", verbose=True)
    logger.info(f"{procs = }")
    assert not procs
    zombies = list_zombies()
    logger.info(f"{zombies = }")
    assert zombies
    assert stub.is_dead_or_zombie()

    time.sleep(0.1)

    # don't know what the return code will be, sometimes its 0, sometimes its 1 or -15
    _ = stub.quit()  # send a SIGTERM to the sub-process

    logger.info(f"{stub.exc_info = }")

    assert not stub.exists()
    assert not stub.is_running()
    assert not is_active(cmd_port)  # this method takes about 1 second because of the timeout (see send() below)
    assert stub.returncode() == 0


def test_process_no_shell():
    stub = SubProcess("Stub Process", [sys.executable, str(find_file("empty_process.py", root=HERE).resolve())])
    cmd_port = 5556

    assert stub.execute()
    assert stub.exc_info == {}
    assert is_active(cmd_port)
    assert not stub.is_dead_or_zombie()
    logger.info(capture_rich_output(list_processes("empty")))
    logger.info(ps_egrep("empty"))
    assert quit_process(cmd_port)


def test_process_with_shell():
    stub = SubProcess(
        "Stub Process",
        [sys.executable, str(find_file("empty_process.py", root=HERE).resolve())],
        shell=True,
    )
    cmd_port = 5556

    assert stub.execute()
    assert stub.exc_info == {}
    assert is_active(cmd_port)
    assert not stub.is_dead_or_zombie()
    logger.info(capture_rich_output(list_processes("empty")))
    logger.info(ps_egrep("empty"))
    assert quit_process(cmd_port)


def test_process_with_children():
    parent = SubProcess(
        "Parent Process",
        [sys.executable, str(find_file("process_with_children.py", root=HERE).resolve())],
    )
    cmd_port = 5557

    assert parent.execute()
    assert parent.exc_info == {}
    assert is_active(cmd_port)
    assert parent.children()
    print("children", parent.children())
    print("list_processes", list_processes("empty"))
    logger.info(ps_egrep("empty"))
    # assert quit_process(cmd_port)
    print("rc = ", parent.quit())
    print("children", parent.children())
    print("list_processes", list_processes("empty"))


def test_raise_value_error():
    proc = SubProcess("ValueErrorApp", [sys.executable, HERE / "scripts" / "raise_value_error.py"])
    proc.execute()

    time.sleep(1.5)  # the script will raise a ValueError after 1.0s

    assert not proc.is_running()
    assert proc.exc_info == {}
    assert proc.is_dead_or_zombie()
    assert proc.returncode() == 1

    proc.quit()


def test_process_status():

    # Need to provide a prefix here otherwise we will get a duplicate metrics error when a control server is created.

    status = ProcessStatus(metrics_prefix="X372")

    sd = status.as_dict()

    print(sd)

    assert sd["PID"] == os.getpid()
    assert "UUID" in sd


# Helper function to communicate with the empty_process.py


def is_active(port: int) -> bool:
    """
    This check is to see if we get a response from the process with ZeroMQ.

    Returns:
        True if process responds to ZeroMQ requests.
    """
    return send(port, "Ping") == "Pong"


def get_status(port: int) -> str:
    """
    Returns status information of the running empty_process.

    Returns:
        ProcessStatus: status information on the running process.
    """
    return send(port, "Status?")


def quit_process(port: int) -> bool:
    """
    Send a Quit command to the sub-process and returns the response.
    """
    return send(port, "Quit") == "Quiting"


def send(port: int, command: str) -> str:
    """
    Sends a command to the sub-process and waits for a reply.

    The command is pickled before sending and the reply is also expected to be pickled.
    If no reply is received after 1 second, None is returned.

    Args:
        port (int): zeromq port where the command should be sent to
        command (str): the command to send to the sub-process

    Returns:
        The unpickled reply from the sub-process.
    """
    reply = None
    logger.info(f"Sending command {command} to {port}")

    pickle_string = pickle.dumps(command)

    with zmq.Context.instance().socket(zmq.REQ) as socket:
        socket.setsockopt(zmq.LINGER, 0)

        socket.connect(f"tcp://localhost:{port}")
        try:
            socket.send(pickle_string, zmq.DONTWAIT)

            if socket.poll(1000, zmq.POLLIN):
                pickle_string = socket.recv(zmq.DONTWAIT)
                reply = pickle.loads(pickle_string)

        except zmq.error as exc:
            logger.exception(f"Send failed with: {exc}")

    return reply



import contextlib
import logging
import re
import sys
import time

import pytest

from egse.dummy import DummyProxy
from egse.process import SubProcess
from egse.proxy import Proxy
from egse.response import Failure
from egse.system import Timer


@contextlib.contextmanager
def dummy_service():

    dummy_dev = SubProcess("Dummy Device", [sys.executable, "-m", "egse.dummy"], ["start-dev"])
    dummy_dev.execute()

    dummy_cs = SubProcess("Dummy CS", [sys.executable, "-m", "egse.dummy"], ["start-cs"])
    dummy_cs.execute()

    processes = [dummy_dev, dummy_cs]

    time.sleep(0.5)  # give the processes the time to start up

    yield processes

    dummy_cs_stop = SubProcess("Dummy CS", [sys.executable, "-m", "egse.dummy"], ["stop-cs"])
    dummy_cs_stop.execute()

    for proc in processes:
        proc.quit()

    time.sleep(0.5)  # give the processes the time to shut down


def is_valid_ip_address_format(string):
    pattern = r"^\d{0,3}\.\d{0,3}\.\d{0,3}\.\d{0,3}$"
    return bool(re.match(pattern, string))


def test_proxy_without_cs():

    proxy = DummyProxy()

    assert proxy.is_cs_connected() is False
    assert proxy.has_commands() is False
    assert proxy.get_commands() == []

    with pytest.raises(NotImplementedError):
        assert proxy.info()


def test_protocol_commands():

    with dummy_service(), DummyProxy() as dummy:

        with Timer("ping", log_level=logging.WARNING, precision=6):
            assert dummy.ping() is True

        assert dummy.get_commanding_port() == 4443
        assert dummy.get_service_port() == 4444
        assert dummy.get_monitoring_port() == 4445

        assert dummy.get_endpoint() == "tcp://localhost:4443"

        assert dummy.has_commands() is True

        for cmd in "info", "get_value", "handle_event":
            assert cmd in dummy.get_commands()

        assert is_valid_ip_address_format(dummy.get_ip_address())

        assert isinstance(dummy.get_service_proxy(), Proxy)
        assert dummy.is_cs_connected() is True


def test_device_commands():

    with dummy_service(), DummyProxy() as dummy:

        with Timer("info", log_level=logging.WARNING, precision=6):
            assert dummy.info().startswith("Dummy Device")

        with Timer("get_value", log_level=logging.WARNING, precision=6):
            assert 0.0 <= dummy.get_value() < 1.0

        assert dummy.division(144, 12) == 12
        response = dummy.division(33, 0)
        assert isinstance(response, Failure)
        assert response.successful is False
        assert response.message == "Executing division failed: : division by zero"
        assert isinstance(response.cause, ZeroDivisionError)



"""
The Storage Control Server, aka Storage Manager, is the service which saves all data coming
from any component in the Common-EGSE.

The Storage manager is implemented as a standard control server.

"""
import logging
import multiprocessing
import sys
from pathlib import Path
from typing import Annotated

import rich
import typer
import zmq
from apscheduler.schedulers.background import BackgroundScheduler
from prometheus_client import start_http_server
from pytz import utc

from egse.control import ControlServer
from egse.env import get_data_storage_location
from egse.env import get_site_id
from egse.process import SubProcess
from egse.settings import Settings
from egse.storage import StorageProtocol
from egse.storage import StorageProxy
from egse.storage import cycle_daily_files

# Use explicit name here otherwise the logger will probably be called __main__

logger = logging.getLogger("egse.storage.storage_cs")

CTRL_SETTINGS = Settings.load("Storage Control Server")
SITE_ID = get_site_id()


class StorageControlServer(ControlServer):
    """
    The Storage Control Server (aka Storage Manager) saves information from registered components.
    """

    def __init__(self):
        super().__init__()

        self.device_protocol = StorageProtocol(self)

        self.logger = logger
        self.logger.debug(f"Binding ZeroMQ socket to {self.device_protocol.get_bind_address()}")

        self.device_protocol.bind(self.dev_ctrl_cmd_sock)

        self.poller.register(self.dev_ctrl_cmd_sock, zmq.POLLIN)

        from egse.confman import ConfigurationManagerProxy, is_configuration_manager_active
        from egse.listener import EVENT_ID

        self.register_as_listener(
            proxy=ConfigurationManagerProxy,
            listener={'name': 'Storage CS', 'proxy': StorageProxy, 'event_id': EVENT_ID.SETUP}
        )

        # NOTE:
        #     Since the CM CS is started after the SM CS in the normal startup sequence, delay the task to load
        #     the Setup until after the CM CS has been properly started. We do that now with a delay time of 30s,
        #     but we might in the future use a function returning a boolean, until...
        self.schedule_task(self.device_protocol.controller.load_setup, after=10.0, when=is_configuration_manager_active)

    def before_serve(self):

        self.scheduler = BackgroundScheduler(timezone=utc)
        self.scheduler.start()
        self.scheduler.add_job(cycle_daily_files, "cron", day="*")

        start_http_server(CTRL_SETTINGS.METRICS_PORT)

    def after_serve(self):
        from egse.confman import ConfigurationManagerProxy

        self.scheduler.shutdown()
        self.unregister_as_listener(proxy=ConfigurationManagerProxy, listener={'name': 'Storage CS'})

    def get_communication_protocol(self):
        return CTRL_SETTINGS.PROTOCOL

    def get_commanding_port(self):
        return CTRL_SETTINGS.COMMANDING_PORT

    def get_service_port(self):
        return CTRL_SETTINGS.SERVICE_PORT

    def get_monitoring_port(self):
        return CTRL_SETTINGS.MONITORING_PORT


app = typer.Typer(name="sm_cs", no_args_is_help=True)


@app.command()
def start():
    """Start the Storage Manager."""

    multiprocessing.current_process().name = "sm_cs"

    # We import this class such that the class name is
    # 'egse.storage.storage_cs.StorageControlServer' and we
    # can compare self with isinstance inside the Control.
    # If this import is not done, the class name for the
    # StorageControlServer would be '__main__.StorageControlServer'.

    from egse.storage.storage_cs import StorageControlServer  # noqa

    try:
        check_prerequisites()
    except RuntimeError as exc:
        logger.info(exc)
        return 0

    try:
        control_server = StorageControlServer()
        control_server.serve()
    except KeyboardInterrupt:
        print("Shutdown requested...exiting")
    except SystemExit as exit_code:
        print(f"System Exit with code {exit_code}.")
        sys.exit(exit_code.code)
    except Exception:
        import traceback

        traceback.print_exc(file=sys.stdout)

    return 0


@app.command()
def start_bg():
    """Start the Storage Manager Control Server in the background."""
    proc = SubProcess("sm_cs", ["sm_cs", "start"])
    proc.execute()


@app.command()
def stop():
    """Send a 'quit_server' command to the Storage Manager."""
    try:
        with StorageProxy() as sm:
            sp = sm.get_service_proxy()
            sp.quit_server()
    except ConnectionError:
        rich.print("[red]ERROR: Couldn't connect to the storage manager.[/]")


@app.command()
def status(full: Annotated[bool, typer.Option(help="Give a full status report")] = False):
    """Print the status of the control server."""

    import rich
    from egse.storage import get_status

    rich.print(get_status(full=full), end='')


def check_prerequisites():
    """Checks if all prerequisites for running the Storage Manager are met.

    Raises:
        RuntimeError when one or more of the prerequisites is not met.
    """

    fails = 0

    # We need a proper location for storing the data, this directory shall contain
    # two subfolders: 'daily' and 'obs'.

    location = get_data_storage_location(site_id=SITE_ID)

    if not location:
        raise RuntimeError(
            "The data storage location is not defined. Please check your environment."
        )

    location = Path(location)

    if not location.exists():
        logger.error(
            f"The directory {location} does not exist, provide a writable location for "
            f"storing the data."
        )
        fails += 1

    logger.debug(f"location = {location}")

    daily_dir = location / "daily"
    obs_dir = location / "obs"

    if not daily_dir.exists():
        logger.error("The data storage location shall have a 'daily' sub-folder.")
        fails += 1
    if not obs_dir.exists():
        logger.error("The data storage location shall have a 'obs' sub-folder.")
        fails += 1

    # now raise the final verdict

    if fails:
        raise RuntimeError(
            "Some of the prerequisites for the Storage Manager haven't met. "
            "Please check the logs."
        )


if __name__ == "__main__":
    sys.exit(app())



import pytest

from egse.device import DeviceError


def test_device_error():
    # DeviceError expects two positional arguments: 'device_name' and 'message'

    with pytest.raises(TypeError):
        raise DeviceError()


def test_device_error_with_args():
    with pytest.raises(DeviceError) as exc:
        raise DeviceError("DAQ6510", "A generic device error")

    assert isinstance(exc.value, DeviceError)
    assert any("generic" in arg for arg in exc.value.args)
    assert any("DAQ" in arg for arg in exc.value.args)
    assert str(exc.value) == "DAQ6510: A generic device error"



import subprocess
import sys
from pathlib import Path

import rich


def start_rs_cs():
    rich.print("Starting the registry service core service...")

    out = open(Path('~/.rs_cs.start.out').expanduser(), 'w')

    subprocess.Popen(
        [sys.executable, '-m', 'egse.registry.server', 'start'],
        stdout=out, stderr=out, stdin=subprocess.DEVNULL,
        close_fds=True
    )


def start_log_cs():
    rich.print("Starting the logging core service...")

    out = open(Path('~/.log_cs.start.out').expanduser(), 'w')

    subprocess.Popen(
        [sys.executable, '-m', 'egse.logger.log_cs', 'start'],
        stdout=out, stderr=out, stdin=subprocess.DEVNULL,
        close_fds=True
    )


def start_sm_cs():
    rich.print("Starting the storage manager core service...")

    out = open(Path('~/.sm_cs.start.out').expanduser(), 'w')

    subprocess.Popen(
        [sys.executable, '-m', 'egse.storage.storage_cs', 'start'],
        stdout=out, stderr=out, stdin=subprocess.DEVNULL,
        close_fds=True
    )


def start_cm_cs():
    rich.print("Starting the configuration manager core service...")

    out = open(Path('~/.cm_cs.start.out').expanduser(), 'w')

    subprocess.Popen(
        [sys.executable, '-m', 'egse.confman.confman_cs', 'start'],
        stdout=out, stderr=out, stdin=subprocess.DEVNULL,
        close_fds=True
    )



"""
This module defines the level, format and handlers for the root logger and for the special
'egse' logger. The egse_logger will be configured with a special handler which sends all
logging messages to a log control server.

This module is loaded whenever an egse module is loaded, to ensure all log messages are properly
forwarded to the log control server.
"""

import logging
import pickle
import sys
import traceback

import zmq

from egse.settings import Settings
from egse.zmq_ser import connect_address

CTRL_SETTINGS = Settings.load("Logging Control Server")

LOG_FORMAT_FULL = (
    "%(asctime)23s:%(processName)20s:%(levelname)8s:%(lineno)5d:%(name)-20s:%(message)s"
)

# Configure the root logger

logging.basicConfig(level=logging.DEBUG, format=LOG_FORMAT_FULL)

__all__ = [
    "egse_logger",
    "set_all_logger_levels",
    "close_all_zmq_handlers",
    "replace_zmq_handler",
    "get_log_file_name",
]


def get_log_file_name():
    """
    Returns the filename of the log file as defined in the Settings or return the default name 'general.log'.
    """
    return CTRL_SETTINGS.get("FILENAME", "general.log")


class ZeroMQHandler(logging.Handler):
    def __init__(self, uri=None, socket_type=zmq.PUSH, ctx=None):

        from egse.settings import Settings
        from egse.zmq_ser import connect_address

        ctrl_settings = Settings.load("Logging Control Server")
        uri = uri or connect_address(ctrl_settings.PROTOCOL, ctrl_settings.HOSTNAME,
                                     ctrl_settings.LOGGING_PORT)

        logging.Handler.__init__(self)

        # print(f"ZeroMQHandler.__init__({uri=}, {socket_type=}, {ctx=})")

        self.setLevel(logging.NOTSET)

        self.ctx = ctx or zmq.Context().instance()
        self.socket = zmq.Socket(self.ctx, socket_type)
        self.socket.setsockopt(zmq.SNDHWM, 0)  # never block on sending msg
        self.socket.connect(uri)

    def __del__(self):
        self.close()

    def close(self):
        self.socket.close(linger=100)

    def emit(self, record):
        """
        Emit a record.

        Writes the LogRecord to the queue, preparing it for pickling first.
        """

        # print(f"ZeroMQHandler.emit({record})")

        from egse.system import is_in_ipython

        try:
            if record.exc_info:
                record.exc_text = traceback.format_exc()
                record.exc_info = None  # traceback objects can not be pickled
            if record.processName == "MainProcess" and is_in_ipython():
                record.processName = "IPython"
            data = pickle.dumps(record.__dict__)
            self.socket.send(data, flags=zmq.NOBLOCK)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as exc:
            print(f"ZeroMQHandler: Exception - {exc}", file=sys.stderr)
            self.handleError(record)


def close_all_zmq_handlers():
    """
    Close all the ZeroMQHandlers that are connected to a logger.

    This function is automatically called upon termination of the control servers. For your own
    applications, call this function before exiting the App.
    """

    loggers = logging.Logger.manager.loggerDict

    for name, logger in loggers.items():
        if isinstance(logger, logging.PlaceHolder):
            continue
        for handler in logger.handlers:
            if isinstance(handler, ZeroMQHandler):
                logger.debug(f"Closing handler for logger {name}")
                handler.close()


# Initialize logging as we want it for the Common-EGSE
#
# * The ZeroMQHandler to send all logging messages, i.e. level=DEBUG to the Logging Server
# * The (local) StreamingHandlers to print only INFO messages and higher

logging.disable(logging.NOTSET)
root_logger = logging.getLogger()

for handler in root_logger.handlers:
    handler.setLevel(logging.INFO)


# Define the `egse` logger and add the ZeroMQHandler to this logger

egse_logger = logging.getLogger("egse")
egse_logger.setLevel(logging.DEBUG)

zmq_handler = ZeroMQHandler()
zmq_handler.setLevel(logging.NOTSET)

egse_logger.addHandler(zmq_handler)
egse_logger.setLevel(logging.DEBUG)


def replace_zmq_handler():
    """
    This function will replace the current ZeroMQ Handler with a new instance. Use this function
    in the run() method of a multiprocessing.Process:

        import egse.logger
        egse.logger.replace_zmq_handler()

    Don't use this function in the __init__() method as only the run() method will execute in
    the new Process and replace the handler in the proper environment. The reason for this is
    that the ZeroMQ socket is not thread/Process safe, so a new ZeroMQ socket needs to be created
    in the correct process environment.
    """
    global egse_logger

    this_handler = None
    for handler in egse_logger.handlers:
        if isinstance(handler, ZeroMQHandler):
            this_handler = handler
    if this_handler is not None:
        egse_logger.removeHandler(this_handler)
    egse_logger.addHandler(ZeroMQHandler())


def create_new_zmq_logger(name: str):
    """
    Create a new logger with the given name and add a ZeroMQ Handler to this logger.

    If the logger already has a ZeroMQ handler attached, don't add a second ZeroMQ handler,
    just return the Logger object.

    Args:
        name: the requested name for the logger

    Returns:
        A Logger for the given name with a ZeroMQ handler attached.
    """
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    # If the ZeroMQ handler already exists for this logger, don't add a second handler

    for handler in logger.handlers:
        if isinstance(handler, ZeroMQHandler):
            return logger

    zmq_handler = ZeroMQHandler()
    zmq_handler.setLevel(logging.NOTSET)

    logger.addHandler(zmq_handler)
    logger.setLevel(logging.DEBUG)

    return logger


def set_all_logger_levels(level: int):
    global root_logger, egse_logger

    root_logger.level = level
    egse_logger.level = level

    for handler in root_logger.handlers:
        handler.setLevel(level)

    # We don't want to restrict egse_logger levels

    # for handler in egse_logger.handlers:
    #     handler.setLevel(level)


TIMEOUT_RECV = 1.0  # seconds


def send_request(command_request: str):
    """Sends a request to the Logger Control Server and waits for a response."""
    ctx = zmq.Context().instance()
    endpoint = connect_address(
        CTRL_SETTINGS.PROTOCOL, CTRL_SETTINGS.HOSTNAME, CTRL_SETTINGS.COMMANDING_PORT
    )
    socket = ctx.socket(zmq.REQ)
    socket.connect(endpoint)

    socket.send(pickle.dumps(command_request))
    rlist, _, _ = zmq.select([socket], [], [], timeout=TIMEOUT_RECV)
    if socket in rlist:
        response = socket.recv()
        response = pickle.loads(response)
    else:
        response = {"error": "Receive from ZeroMQ socket timed out for Logger Control Server."}
    socket.close(linger=0)

    return response


if __name__ == "__main__":

    import egse.logger

    LOGGER = logging.getLogger("egse.0mq-log-test")
    LOGGER.debug("Hello, ZeroMQ logging: This is a DEBUG message.")
    LOGGER.info("Hello, ZeroMQ logging: This is an INFO message.")
    LOGGER.warning("Hello, ZeroMQ logging: This is a WARNING message.")
    LOGGER.error("Hello, ZeroMQ logging: This is an ERROR message.")
    LOGGER.critical("Hello, ZeroMQ logging: This is a CRITICAL message.")
    try:
        raise ValueError("A fake ValueError, raised for testing.")
    except ValueError:
        LOGGER.exception("Hello, ZeroMQ logging: This is an EXCEPTION message.")

    LOGGER = logging.getLogger("plain-log-test")

    LOGGER.debug("Vanilla logging: This is a DEBUG message.")
    LOGGER.info("Vanilla logging: This is an INFO message.")
    LOGGER.warning("Vanilla logging: This is a WARNING message.")
    LOGGER.error("Vanilla logging: This is an ERROR message.")
    LOGGER.critical("Vanilla logging: This is a CRITICAL message.")
    try:
        raise ValueError("A fake ValueError, raised for testing.")
    except ValueError:
        LOGGER.exception("Vanilla logging: This is an EXCEPTION message.")



"""
The Process Manager Control Server is the service that handles all Common-EGSE
processes.  The list of processes to handle, is taken from the Configuration
Manager (from the setup).

The Process Manager Control Server is implemented as a standard control server.
"""
import logging
import multiprocessing
import sys

import click
import zmq
from prometheus_client import start_http_server

from egse.control import ControlServer
from egse.procman import ProcessManagerProtocol
from egse.procman import ProcessManagerProxy
from egse.procman import is_process_manager_cs_active
from egse.settings import Settings

CTRL_SETTINGS = Settings.load("Process Manager Control Server")

logger = logging.getLogger(__name__)


class ProcessManagerControlServer(ControlServer):
    """
    The Process Manager Control Server handles all Common-EGSE processes.
    """

    def __init__(self):
        """Initialisation of a new Control Server for Process Management.

        The initialisation of this Control Server consists of the following
        steps:

            - Define the device protocol;
            - Bind the command socket to the device protocol (to listen for
              commands);
            - Add the device protocol to the poller, to be able to listen for
              commands on different sockets in one thread.
        """

        super().__init__()

        # Protocol used for commanding

        self.device_protocol = ProcessManagerProtocol(self)

        # Bind to a socket to listen for commands

        bind_address = self.device_protocol.get_bind_address()

        self.logger = logger
        self.logger.debug(f"Binding ZeroMQ socket to {bind_address}")

        self.device_protocol.bind(self.dev_ctrl_cmd_sock)

        # Listen on different sockets in the same threads -> poller

        self.poller.register(self.dev_ctrl_cmd_sock, zmq.POLLIN)

    def get_communication_protocol(self):
        """Returns the communication protocol, as defined in the settings.

        Returns:
            - Communication protocol, as defined in the settings.
        """
        return CTRL_SETTINGS.PROTOCOL

    def get_commanding_port(self):
        """Returns the commanding port number.

        The commanding port is the port on which the Controller listens for
        commands, using the REQ-REP (ZeroMQ) socket pattern.  Its number is
        read from the settings file.

        Returns:
            - Number of the port on which the Controller listens for commands.
        """

        return CTRL_SETTINGS.COMMANDING_PORT

    def get_service_port(self):
        """Returns the service port number.

        The service port is the port on which the Controller listens for
        configuration and administration, using the PUB-SUB (ZeroMQ) socket
        pattern.  Its number is read from the settings file.

        Returns:
            - Number of the port on which the Controller listens for
              configuration and administration.
        """

        return CTRL_SETTINGS.SERVICE_PORT

    def get_monitoring_port(self):
        """Returns the monitoring port number.

        The monitoring port is the port on which the Controller sends periodic
        information on the device, using the PUB-SUB (ZeroMQ) socket
        pattern.  Its number is read from the settings file.

        Returns:
            - Number of the port on which the Controller sends periodic status
              information on the device.
        """

        return CTRL_SETTINGS.MONITORING_PORT

    def get_storage_mnemonic(self):
        """Returns the storage mnemonic for the Controller.

        The storage mnemonic is used in the filename of the housekeeping of
        the Controller (as using by the Storage).  If this is not defined in
        the settings file, "PM" will be used instead.

        Returns:
            - Storage mnemonic for the Controller.
        """

        try:

            # As defined in the settings file

            return CTRL_SETTINGS.STORAGE_MNEMONIC

        except AttributeError:

            # Default

            return "PM"

    def before_serve(self):
        start_http_server(CTRL_SETTINGS.METRICS_PORT)


@click.group()
def cli():
    pass


@cli.command()
def start():
    """Start the Process Manager."""

    multiprocessing.current_process().name = "pm_cs"

    # We import this class such that the class name is
    # 'egse.procman.procman_cs.ProcessManagerControlServer' and we
    # can compare self with isinstance inside the Control.
    # If this import would not be done, the class name for the
    # ProcessManagerControlServer would be '__main__.ProcessManagerControlServer'.

    from egse.procman.procman_cs import ProcessManagerControlServer  # noqa

    try:
        control_server = ProcessManagerControlServer()
        control_server.serve()
    except KeyboardInterrupt:
        print("Shutdown requested...exiting")
    except SystemExit as exit_code:
        print("System Exit with code {}.".format(exit_code))
        sys.exit(exit_code)
    except Exception:
        import traceback

        traceback.print_exc(file=sys.stdout)

    return 0


@cli.command()
def stop():
    """Send a 'quit_server' command to the Process Manager."""
    with ProcessManagerProxy() as pm:
        sp = pm.get_service_proxy()
        sp.quit_server()


@cli.command()
def status():
    """Send a 'quit_server' command to the Process Manager."""

    import rich
    rich.print("Process Manager:")
    if is_process_manager_cs_active():
        rich.print(f"  Status: [green]active")
        with ProcessManagerProxy() as pm:
            rich.print(f"  Hostname: {pm.get_ip_address()}")
            rich.print(f"  Monitoring port: {pm.get_monitoring_port()}")
            rich.print(f"  Commanding port: {pm.get_commanding_port()}")
            rich.print(f"  Service port: {pm.get_service_port()}")
    else:
        rich.print(f"  Status: [red]not active")


if __name__ == "__main__":
    sys.exit(cli())



from dataclasses import dataclass

import pytest

from egse.env import set_default_environment
from fixtures.helpers import setup_conf_data
from fixtures.helpers import setup_data_storage_layout
from fixtures.helpers import teardown_conf_data
from fixtures.helpers import teardown_data_storage_layout


@dataclass
class DefaultEnvironment:
    project: str
    site_id: str
    data_root: str


@pytest.fixture(scope="session", autouse=True)
def default_env(tmp_path_factory):

    project = "CGSE"
    site_id = "LAB23"

    tmp_data_dir = tmp_path_factory.mktemp("data")

    set_default_environment(project, site_id, tmp_data_dir)

    data_root = setup_data_storage_layout(tmp_data_dir)
    setup_conf_data(tmp_data_dir)

    from egse.env import print_env
    print_env()

    yield DefaultEnvironment(project=project, site_id=site_id, data_root=str(data_root))

    teardown_conf_data(tmp_data_dir)
    teardown_data_storage_layout(tmp_data_dir)



import logging
import os
import signal
import sys

import typer

from egse.system import SignalCatcher

app = typer.Typer()
logger = logging.getLogger("egse.test.process")


@app.command()
def main(ignore_sigterm: bool = False):
    logging.basicConfig(level=logging.INFO)

    logger.info(f"{sys.version_info = }")
    logger.info(f"{ignore_sigterm = }")

    if ignore_sigterm:
        rc = _ignore_sigterm()
    else:
        rc = _handle_sigterm()

    raise typer.Exit(code=rc)


def _handle_sigterm():
    killer = SignalCatcher()

    while "waiting for a SIGTERM signal":
        if killer.term_signal_received:
            logger.info(f"{killer.signal_number = }")
            if killer.signal_number == signal.SIGTERM:
                logger.info("SIGTERM received, terminating...")
                return 42

    # The following code will never execute
    return 0


def _ignore_sigterm():
    killer = SignalCatcher()

    while "ignoring a SIGTERM signal":
        if killer.term_signal_received:
            logger.info(f"{killer.signal_number = }")
            if killer.signal_number == signal.SIGTERM:
                logger.info("SIGTERM received and ignored.")
                killer.clear(term=True)
                continue

    # The following code will never execute
    return 0


if __name__ == "__main__":
    print(f"{sys.argv=}")
    print(f"pid={os.getpid()}")

    try:
        from egse.logger import replace_zmq_handler
    except (ModuleNotFoundError, ImportError):
        def replace_zmq_handler():
            ...

    try:
        app()
    except typer.Exit as e:
        raise SystemExit(e.exit_code)



import pytest

from egse.coordinates.point import Point
from egse.coordinates.referenceFrame import ReferenceFrame


def test_construction():
    master = ReferenceFrame.createMaster()

    p1 = Point([1, 2, 3], ref=master, name="P1")
    p2 = Point([1, 2, 3], ref=master, name="P2")
    assert p1 is not p2
    assert p1 == p2

    p3 = Point([1, 2, 3], ref=master, name="P1")
    assert p3 is not p1
    assert p3 == p1 == p2


def test_equal():
    master = ReferenceFrame.createMaster()

    p1 = Point([1, 2, 3], ref=master)
    p2 = Point([1, 2, 3], ref=master)

    assert p1 == p1
    assert p1 != [1, 2, 3]

    assert p1.name != p2.name
    assert p1 is not p2
    assert p1 == p2

    p3 = Point([1, 3, 5], ref=master)

    assert p3 is not p2
    assert p3 != p2

    r1 = ReferenceFrame.fromTranslation(0., 1., 2., ref=master)
    p4 = Point([1, 3, 5], ref=r1)

    assert p4 is not p3
    assert p4 != p3


def test_isSame():
    master = ReferenceFrame.createMaster()

    p1 = Point([1, 2, 3], ref=master)
    p2 = Point([1, 2, 3], ref=master)

    assert p1.isSame(p2)

    r1 = ReferenceFrame.fromTranslation(1., 2., 3., ref=master)
    p3 = Point([1, 3, 5], ref=master)
    p4 = Point([0, 0, 0], ref=r1)

    assert p4.isSame(p2)
    assert p2.isSame(p4)
    assert not p4.isSame(p3)
    assert not p3.isSame(p4)


def test_addition():
    master = ReferenceFrame.createMaster()

    p1 = Point([1, 2, 3], ref=master, name="P1")
    p2 = Point([1, 2, 3], ref=master, name="P2")

    assert p1 + p2 == Point([2, 4, 6], ref=master)

    r1 = ReferenceFrame.fromTranslation(0, 0, 3, ref=master)
    p3 = Point([1, 2, 0], ref=r1)

    with pytest.raises(TypeError):
        assert p1 + p3 == Point([2, 4, 6], ref=master)


def test_subtraction():
    master = ReferenceFrame.createMaster()

    p1 = Point([1, 2, 3], ref=master, name="P1")
    p2 = Point([1, 2, 3], ref=master, name="P2")

    assert p1 - p2 == Point([0, 0, 0], ref=master)

    r1 = ReferenceFrame.fromTranslation(0, 0, 3, ref=master)
    p3 = Point([1, 2, 0], ref=r1)

    with pytest.raises(TypeError):
        assert p1 - p3 == Point([0, 0, 0], ref=master)



import sys
import time
from typing import List

import pytest
import rich

from egse.confman import ConfigurationManagerProxy
from egse.confman import is_configuration_manager_active
from egse.confman.confman_cs import list_setups
from egse.process import SubProcess


def test_is_cm_cs_is_active():

    assert is_configuration_manager_active() in (False, True)  # Should not raise an exception


@pytest.mark.skipif(not is_configuration_manager_active(), reason="core service cm_cs not running")
def test_list_setups():

    rich.print()

    with ConfigurationManagerProxy() as cm:
        setups = cm.list_setups()

    assert isinstance(setups, List)

    # FIXME: This check is dependent on the current environment that was set up to run the core services

    assert setups[0] == ('00000', 'VACUUM_LAB', 'Initial zero Setup for VACUUM_LAB', 'no sut_id')


@pytest.mark.skipif(not is_configuration_manager_active(), reason="core service cm_cs not running")
def test_load_setup():

    with ConfigurationManagerProxy() as cm:
        setup = cm.load_setup(setup_id=0)
        assert setup.get_id() == '00000'

        # load_setup(..) does change the Setup that is loaded on the cm_cs

        setup = cm.load_setup(setup_id=1)
        assert setup.get_id() == '00001'

        setup = cm.get_setup()
        assert setup.get_id() == '00001'


@pytest.mark.skipif(not is_configuration_manager_active(), reason="core service cm_cs not running")
def test_get_setup():

    with ConfigurationManagerProxy() as cm:
        setup = cm.load_setup(setup_id=0)
        assert setup.get_id() == '00000'

        # get_setup(..) doesn't change the Setup that is loaded on the cm_cs

        setup = cm.get_setup(setup_id=1)
        assert setup.get_id() == '00001'

        setup = cm.get_setup()
        assert setup.get_id() == '00000'


@pytest.mark.skipif(not is_configuration_manager_active(), reason="core service cm_cs not running")
def test_listeners():

    dummy_dev = SubProcess("Dummy Device", [sys.executable, "-m", "egse.dummy", "start-dev"])
    dummy_dev.execute()

    dummy_cs = SubProcess("Dummy CS", [sys.executable, "-m", "egse.dummy", "start-cs"])

    try:
        with ConfigurationManagerProxy() as cm:

            assert 'Dummy CS' not in cm.get_listener_names()

            dummy_cs.execute()

            time.sleep(0.5)  # Registration needs some time

            assert 'Dummy CS' in cm.get_listener_names()

            cm.load_setup(setup_id=1)

    finally:
        dummy_dev.quit()
        dummy_cs.quit()

        while dummy_dev.is_running():
            time.sleep(1.0)



"""
This module provides convenience functions to use resources in your code without
the need to specify an absolute path or try to locate the resources in your local
installation (which is time-consuming, error-prone, and introduces quite some
redundancy).

Resources can be files of different format that are distributed together with the
source code, e.g.

    * image data
    * icons
    * YAML files
    * binary files, e.g. dynamic libraries
    * style files for GUI applications
    * calibration files distributed with the source code

Each of the resources have a fixed location within the source tree and is identified
with a resource identifier. There are a number of default identifier that are defined
as follows:

    * `icons`: located in the sub-folder 'icons'
    * `images`: located in the sub-folder 'images'
    * `styles`: located in the sub-folder 'data'
    * `data`: located in the sub-folder 'data'
    * `lib`: located in sub-folder 'lib'

Resource shall be initialised by either the process or the library using the function
`initialise_resources()`. The function optionally takes a locations in which it will
search for the default resource ids. The usual way to initialise is:

    >>> initialise_resources(Path(__file__).parent)

Another way to make your resource available is through entry points in the
distribution of your package. In the `pyproject.toml` file, you can specify resources
as follows:

    [project.entry-points."cgse.resource"]
    icons = 'egse.gui.icons'
    styles = 'egse.gui.styles'

This will automatically add these resources during initialisation.

Resources can be accessed from the code without specifying the absolute pathname,
using a `:/resource_id/` that is known by the resource module. A wildcard can be
introduced after the `resource_id` to indicate the resource is in one of the
subdirectories.

Example usage:
    * get_resource(":/icons/open-document.png")
    * get_resource(":/styles/dark.qss")
    * get_resource(":/lib/*/EtherSpaceLink_v34_86.dylib")

A new `resource_id` can be added with the `add_resource_id()` function, specifying a
resource_id string and a location. That location will be added to the list of locations
for that resource id. To add a resource location for e.g. configuration files, do

    >>> from egse.resource import add_resource_id
    >>> add_resource_id('styles', Path(__file__).parent)

Alternatives

The `egse.config` module has a number of alternatives for locating files and resources.

    * find_file(..) and find_files(..)
    * find_dir(..) and find_dirs(..)
    * get_resource_dirs()
    * get_resource_path()

The functions for finding files and directories are more flexible, but take more
time and effort. They are mainly used for dynamically searching for a file or
folder, not necessarily within the source code location.

The resource specific functions in the egse.config module will be deprecated when
their functionality is fully replaced by this `egse.resource` module.

"""

__all__ = [
    "AmbiguityError",
    "NoSuchFileError",
    "ResourceError",
    "add_resource_id",
    "get_resource",
    "get_resource_dirs",
    "get_resource_locations",
    "get_resource_path",
    "initialise_resources",
    "print_resources",
]

import contextlib
import errno
import logging
import re
from pathlib import Path
from pathlib import PurePath
from typing import Dict
from typing import List
from typing import Union

from os.path import exists
from os.path import join

from egse.config import find_first_occurrence_of_dir
from egse.config import find_files
from egse.exceptions import InternalError
from egse.plugin import entry_points
from egse.system import get_package_location

_LOGGER = logging.getLogger(__name__)


class ResourceError(Exception):
    """Base class, raised when a resource is not defined."""


class AmbiguityError(ResourceError):
    """Raised when more than one option is possible."""


class NoSuchFileError(ResourceError):
    """Raised when no file could be found for the given resource."""


# Testing regex: https://pythex.org

PATTERN = re.compile(r"^:/(\w+)/(\*+/)?(.*)$")

# Default resources will be checked when executing the initialise_resources()
# optionally with a path as root. It is not needed to use add_resource_id()
# for these resources.

DEFAULT_RESOURCES = {
    "icons": "/icons",
    "images": "/images",
    "styles": "/styles",
    "lib": "/lib",
    "data": "/data",
}

_RESOURCE_DIRS = ["resources", "icons", "images", "styles", "data"]

resources = {}


def check_if_file_exists(filename: Union[Path, str], resource_id: str = None) -> Path:
    """
    Check if the given filename exists. If the filename exists, return the filename, else raise a
    NoSuchFileError.

    Args:
        filename (Path|str): an absolute filename
        resource_id (str): a resource identifier

    Return:
        The given filename if it exists.

    Raises:
        NoSuchFileError if the given filename doesn't exist.
    """
    filename = Path(filename)
    if filename.is_file():
        return filename

    if resource_id:
        raise NoSuchFileError(f"The file '{filename.name}' could not be found for the given resource '{resource_id}'")
    else:
        raise NoSuchFileError(f"The file '{filename.name}' doesn't exist.")


def contains_wildcard(filename: str):
    """
    Returns True if the filename contains a wildcard, otherwise False.
    A wildcard is an asterisk '*' or a question mark '?' character.
    """
    if "*" in filename:
        return True
    if "?" in filename:
        return True

    return False


def get_resource_locations() -> Dict[str, List[Path]]:
    """
    Returns a dictionary of names that can be used as resource location.
    The keys are strings that are recognised as valid resource identifiers, the
    values are a list of their actual absolute path names.
    """
    return resources.copy()


def get_resource_dirs(root_dir: Union[str, PurePath]) -> List[Path]:
    """
    Define directories that contain resources like images, icons, and data files.

    Resource directories can have the following names: `resources`, `data`, `icons`, or `images`.
    This function checks if any of the resource directories exist in the `root_dir` that is given as an argument.

    For all existing directories the function returns the absolute path.

    If the argument root_dir is None, an empty list will be returned and a warning message will be issued.

    Args:
        root_dir (str): the directory to search for resource folders

    Returns:
        a list of absolute Paths.
    """

    if root_dir is None:
        _LOGGER.warning("The argument root_dir can not be None, an empty list is returned.")
        return []

    root_dir = Path(root_dir).resolve()
    if not root_dir.is_dir():
        root_dir = root_dir.parent

    result = []
    for dir_ in _RESOURCE_DIRS:
        if (root_dir / dir_).is_dir():
            result.append(Path(root_dir, dir_).resolve())

    return result


def get_resource_path(name: str, resource_root_dir: Union[str, PurePath] = None) -> PurePath:
    """
    Searches for a data file (resource) with the given name.

    When `resource_root_dir` is not given, the search for resources will start at the root
    folder of the project (using the function `get_common_egse_root()`). Any other root
    directory can be given, e.g. if you want to start the search from the location of your
    source code file, use `Path(__file__).parent` as the `resource_root_dir` argument.

    Args:
        name (str): the name of the resource that is requested
        resource_root_dir (str): the root directory w_HERE the search for resources should be started

    Returns:
        the absolute path of the data file with the given name. The first name that matches
        is returned. If no file with the given name or path exists, a FileNotFoundError is raised.

    """
    for resource_dir in get_resource_dirs(resource_root_dir):
        resource_path = join(resource_dir, name)
        if exists(resource_path):
            return Path(resource_path).absolute()
    raise FileNotFoundError(errno.ENOENT, f"Could not locate resource '{name}'")


def initialise_resources(root: Union[Path, str] = Path(__file__).parent):
    """
    Initialise the default resources and any resource published by a package entry point.

    The argument `root` specifies the root location for the resources. If not specified,
    the location of this module is taken as the root location. So, if you have installed
    this package with `pip install`, you should give the location of your project's source
    code as the root argument.

    When you have specified entry points for the group 'cgse.resource' in your project,
    these resources will also be initialised. Check the global documentation of this module
    for an example entry point.

    Args:
        root (Path|str): the root location for the resources.

    Returns:
        None.
    """

    #  the resources with their absolute path names

    for resource_id in DEFAULT_RESOURCES:
        folder = find_first_occurrence_of_dir(DEFAULT_RESOURCES[resource_id], root=root)
        if folder is not None:
            x = resources.setdefault(resource_id, [])
            if folder not in x:
                x.append(folder)

    for ep in entry_points("cgse.resource"):
        for location in get_package_location(ep.value):
            x = resources.setdefault(ep.name, [])
            if location not in x:
                x.append(location)

    _LOGGER.debug(f"Resources have been initialised: {resources = }")


def print_resources():
    """Prints the currently defined resources."""

    if resources:
        print("Available resources:")
    else:
        print("No resources defined.")
        return

    for resource_id, locations in resources.items():
        print(f"  {resource_id}:")
        for location in locations:
            print(f"    {location}")


def add_resource_id(resource_id: str, location: Union[Path, str]):
    """
    Adds a resource identifier with the given location. Resources can then be specified
    using this resource id.

    The location can be an absolute or relative pathname. In the latter case the path
    will be expanded against the current working directory.

    Args:
        resource_id (str): a resource identifier
        location (Path|str): an absolute or relative pathname

    Returns:
        ValueError if the location can not be determined or is not a directory.
    """

    # Check if location exists and is a directory.

    location = Path(location).expanduser().resolve()

    if not location.exists():
        raise ValueError(f"Unknown location '{location}'")

    if location.is_dir():
        x = resources.setdefault(resource_id, [])
        if location not in x:
            x.append(location)
    else:
        raise ValueError(f"Location is not a directory: {location=}")


def get_resource(resource_locator: str) -> Path:
    """
    Returns the absolute Path for the given resource_locator. The resource_locator consists of
    a resource_id, an optional wildcard and a filename separated by a forward slash '/' and
    started by a colon ':'.

        ':/<resource_id>/[*/]<filename>'

    If the resource_locator starts with a colon ':', the name will be interpreted as a resource_id
    and filename combination and parsed as such.

    If the resource_locator doesn't start with a colon ':', then the string will be interpreted as a
    Path name and returned if that path exists, otherwise a ResourceError is raised.

    The filename can contain the wildcard '*' and/or '?', however the use of a wildcard in the
    filename can still only match one unique filename. This can be useful e.g. if you know the
    filename except for one part of it like a timestamp. Used, e.g., for matching Setup files which
    are unique filenames with a timestamp.

    Args:
        resource_locator (str): a special resource name or a filename

    Returns:
        a Path with the absolute filename for the resource.

    Raises:
        ResourceError when no file could be found or the search is ambiguous.

    """
    # Try to match the special resource syntax `:/resource_id/` or `:/resource_id/*/`

    if resource_locator.startswith(":"):
        match = PATTERN.fullmatch(resource_locator)
        resource_id = match[1]
        filename = match[3]
        try:
            resource_locations = resources[resource_id]
        except KeyError:
            raise ResourceError(f"Resource not defined: {resource_id}")

        # match[2] can be only three things
        #   - None in which case the file must be in the resource location directly
        #   - '*/' in which case the file must be in a sub-folder of the resource
        #   - '**/' to find the file in any sub-folder below the given resource

        if match[2] is None:
            # This will return the first occurrence of the filename

            for resource_location in resource_locations:
                if contains_wildcard(filename):
                    files = list(find_files(filename, root=resource_location))

                    if len(files) == 0:
                        continue
                    elif len(files) == 1:
                        return files[0]
                    else:
                        raise AmbiguityError(
                            f"The {filename=} found {len(files)} matches for the given resource '{resource_id}'."
                        )

                with contextlib.suppress(NoSuchFileError):
                    return check_if_file_exists(resource_location / filename, resource_id)
            else:
                raise NoSuchFileError(f"No file found that matches {filename=} for the given resource '{resource_id}'.")

        elif match[2] == "*/":
            # This will return the first occurrence of the filename

            for resource_location in resource_locations:
                files = list(find_files(filename, root=resource_location))

                if len(files) == 0:
                    continue
                elif len(files) == 1:
                    return files[0]
                else:
                    raise AmbiguityError(
                        f"The {filename=} found {len(files)} matches for the given resource '{resource_id}'."
                    )

            else:
                raise NoSuchFileError(f"No file found that matches {filename=} for the given resource '{resource_id}'.")

        elif match[2] == "**/":
            for resource_location in resource_locations:
                files = list(find_files(filename, root=resource_location))

                if len(files) == 0:
                    continue
                elif len(files) == 1:
                    return files[0]
                else:
                    raise AmbiguityError(
                        f"The {filename=} found {len(files)} matches for the given resource '{resource_id}'."
                    )

            else:
                raise NoSuchFileError(f"No file found that matches {filename=} for the given resource '{resource_id}'.")

        else:
            raise InternalError(f"This shouldn't happen, the match is {match[2]=} for {resource_locator=}")
    else:
        return check_if_file_exists(Path(resource_locator))


# Now initialise the resources, this will
#
#   * Add resource locations in this project (cgse-core) for the default resource ids
#   * Add any other resource locations for the rentry points 'cgse.resource'

initialise_resources()


if __name__ == "__main__":
    import rich

    rich.print("Default resources:")
    rich.print(get_resource_locations())



import logging
import socket
import time

from egse.command import ClientServerCommand
from egse.device import DeviceConnectionError
from egse.device import DeviceConnectionInterface
from egse.device import DeviceError
from egse.device import DeviceTimeoutError
from egse.device import DeviceTransport
from egse.settings import Settings
from egse.system import Timer

logger = logging.getLogger(__name__)

IDENTIFICATION_QUERY = "*IDN?"

DEVICE_SETTINGS = Settings.load("Keithley DAQ6510")
DEVICE_NAME = "DAQ6510"
READ_TIMEOUT = DEVICE_SETTINGS.TIMEOUT  # [s], can be smaller than timeout (for DAQ6510Proxy) (e.g. 1s)


class DAQ6510Command(ClientServerCommand):
    def get_cmd_string(self, *args, **kwargs) -> str:
        """Constructs the command string, based on the given positional and/or keyword arguments.

        Args:
            *args: Positional arguments that are needed to construct the command string
            **kwargs: Keyword arguments that are needed to construct the command string

        Returns: Command string with the given positional and/or keyword arguments filled out.
        """

        out = super().get_cmd_string(*args, **kwargs)
        return out + "\n"


class DAQ6510EthernetInterface(DeviceConnectionInterface, DeviceTransport):
    """Defines the low-level interface to the Keithley DAQ6510 Controller."""

    def __init__(self, hostname: str = None, port: int = None):
        """Initialisation of an Ethernet interface for the DAQ6510.

        Args:
            hostname(str): Hostname to which to open a socket
            port (int): Port to which to open a socket
        """

        super().__init__()

        self.hostname = DEVICE_SETTINGS.HOSTNAME if hostname is None else hostname
        self.port = DEVICE_SETTINGS.PORT if port is None else port
        self._sock = None

        self._is_connection_open = False

    def connect(self) -> None:
        """Connects the device.

        Raises:
            DeviceConnectionError: When the connection could not be established. Check the logging messages for more
                                   details.
            DeviceTimeoutError: When the connection timed out.
            ValueError: When hostname or port number are not provided.
        """

        # Sanity checks

        if self._is_connection_open:
            logger.warning(f"{DEVICE_NAME}: trying to connect to an already connected socket.")
            return

        if self.hostname in (None, ""):
            raise ValueError(f"{DEVICE_NAME}: hostname is not initialized.")

        if self.port in (None, 0):
            raise ValueError(f"{DEVICE_NAME}: port number is not initialized.")

        # Create a new socket instance

        try:
            self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            # The following lines are to experiment with blocking and timeout, but there is no need.
            # self._sock.setblocking(1)
            # self._sock.settimeout(3)
        except socket.error as e_socket:
            raise DeviceConnectionError(DEVICE_NAME, "Failed to create socket.") from e_socket

        # Attempt to establish a connection to the remote host

        # FIXME: Socket shall be closed on exception?

        # We set a timeout of 3s before connecting and reset to None (=blocking) after the `connect` method has been
        # called. This is because when no device is available, e.g. during testing, the timeout will take about
        # two minutes, which is way too long. It needs to be evaluated if this approach is acceptable and not causing
        # problems during production.

        try:
            logger.debug(f'Connecting a socket to host "{self.hostname}" using port {self.port}')
            self._sock.settimeout(3)
            self._sock.connect((self.hostname, self.port))
            self._sock.settimeout(None)
        except ConnectionRefusedError as exc:
            raise DeviceConnectionError(DEVICE_NAME, f"Connection refused to {self.hostname}:{self.port}.") from exc
        except TimeoutError as exc:
            raise DeviceTimeoutError(DEVICE_NAME, f"Connection to {self.hostname}:{self.port} timed out.") from exc
        except socket.gaierror as exc:
            raise DeviceConnectionError(DEVICE_NAME, f"Socket address info error for {self.hostname}") from exc
        except socket.herror as exc:
            raise DeviceConnectionError(DEVICE_NAME, f"Socket host address error for {self.hostname}") from exc
        except socket.timeout as exc:
            raise DeviceTimeoutError(DEVICE_NAME, f"Socket timeout error for {self.hostname}:{self.port}") from exc
        except OSError as exc:
            raise DeviceConnectionError(DEVICE_NAME, f"OSError caught ({exc}).") from exc

        self._is_connection_open = True

        # Check that we are connected to the controller by issuing the "VERSION" or
        # "*ISDN?" query. If we don't get the right response, then disconnect automatically.

        if not self.is_connected():
            raise DeviceConnectionError(DEVICE_NAME, "Device is not connected, check logging messages for the cause.")

    def disconnect(self) -> None:
        """Disconnects from the Ethernet connection.

        Raises:
            DeviceConnectionError when the socket could not be closed.
        """

        try:
            if self._is_connection_open:
                logger.debug(f"Disconnecting from {self.hostname}")
                self._sock.close()
                self._is_connection_open = False
        except Exception as e_exc:
            raise DeviceConnectionError(DEVICE_NAME, f"Could not close socket to {self.hostname}") from e_exc

    def reconnect(self):
        """Reconnects to the device controller.

        Raises:
            ConnectionError when the device cannot be reconnected for some reason.
        """

        if self._is_connection_open:
            self.disconnect()
        self.connect()

    def is_connected(self) -> bool:
        """Checks if the device is connected.

        This will send a query for the device identification and validate the answer.

        Returns: True is the device is connected and answered with the proper ID; False otherwise.
        """

        if not self._is_connection_open:
            return False

        try:
            version = self.query(IDENTIFICATION_QUERY)
        except DeviceError as exc:
            logger.exception(exc)
            logger.error("Most probably the client connection was closed. Disconnecting...")
            self.disconnect()
            return False

        if "DAQ6510" not in version:
            logger.error(
                f'Device did not respond correctly to a "VERSION" command, response={version}. Disconnecting...'
            )
            self.disconnect()
            return False

        return True

    def write(self, command: str) -> None:
        """Senda a single command to the device controller without waiting for a response.

        Args:
            command (str): Command to send to the controller

        Raises:
            DeviceConnectionError when the command could not be sent due to a communication problem.
            DeviceTimeoutError when the command could not be sent due to a timeout.
        """

        try:
            command += "\n" if not command.endswith("\n") else ""

            self._sock.sendall(command.encode())

        except socket.timeout as e_timeout:
            raise DeviceTimeoutError(DEVICE_NAME, "Socket timeout error") from e_timeout
        except socket.error as e_socket:
            # Interpret any socket-related error as a connection error
            raise DeviceConnectionError(DEVICE_NAME, "Socket communication error.") from e_socket
        except AttributeError:
            if not self._is_connection_open:
                msg = "The DAQ6510 is not connected, use the connect() method."
                raise DeviceConnectionError(DEVICE_NAME, msg)
            raise

    def trans(self, command: str) -> str:
        """Sends a single command to the device controller and block until a response from the controller.

        This is seen as a transaction.

        Args:
            command (str): Command to send to the controller

        Returns:
            Either a string returned by the controller (on success), or an error message (on failure).

        Raises:
            DeviceConnectionError when there was an I/O problem during communication with the controller.
            DeviceTimeoutError when there was a timeout in either sending the command or receiving the response.
        """

        try:
            # Attempt to send the complete command

            command += "\n" if not command.endswith("\n") else ""

            self._sock.sendall(command.encode())

            # wait for, read and return the response from HUBER (will be at most TBD chars)

            return_string = self.read()

            return return_string.decode().rstrip()

        except socket.timeout as e_timeout:
            raise DeviceTimeoutError(DEVICE_NAME, "Socket timeout error") from e_timeout
        except socket.error as e_socket:
            # Interpret any socket-related error as an I/O error
            raise DeviceConnectionError(DEVICE_NAME, "Socket communication error.") from e_socket
        except ConnectionError as exc:
            raise DeviceConnectionError(DEVICE_NAME, "Connection error.") from exc
        except AttributeError:
            if not self._is_connection_open:
                raise DeviceConnectionError(DEVICE_NAME, "Device not connected, use the connect() method.")
            raise

    def read(self) -> bytes:
        """Reads from the device buffer.

        Returns: Content of the device buffer.
        """

        n_total = 0
        buf_size = 2048

        # Set a timeout of READ_TIMEOUT to the socket.recv

        saved_timeout = self._sock.gettimeout()
        self._sock.settimeout(READ_TIMEOUT)

        try:
            for idx in range(100):
                time.sleep(0.001)  # Give the device time to fill the buffer
                data = self._sock.recv(buf_size)
                n = len(data)
                n_total += n
                if n < buf_size:
                    break
        except socket.timeout:
            logger.warning(f"Socket timeout error for {self.hostname}:{self.port}")
            return b"\r\n"
        except TimeoutError as exc:
            logger.warning(f"Socket timeout error: {exc}")
            return b"\r\n"
        finally:
            self._sock.settimeout(saved_timeout)

        # logger.debug(f"Total number of bytes received is {n_total}, idx={idx}")

        return data


def main():
    return 0


if __name__ == '__main__':
    main()
